from __future__ import print_function

from abopt.algs.trustregion import cg_steihaug, TrustRegionCG

import numpy
from numpy.testing import assert_allclose
from abopt.testing import RosenProblem, ChiSquareProblem

import pytest
def test_cg_steihaug():
    import numpy
    #Hessian = numpy.diag([1, 2, 3, 400.**2])
    J = numpy.array([[0, 0, 0, 1],
                      [0, 0, 2, 0], 
                      [0, 3, 0, 0], 
                      [400, 0, 0, 0]])
    C = numpy.array([[2, 0, 0, 0],
                      [0, 1, 0, 0], 
                      [0, 0, 3, 0], 
                      [0, 0, 0, 1]])

    problem = ChiSquareProblem(J=J)

    g = numpy.zeros(4) + 1.0
    g[...] = [  -2.,   -4.,   -6., -800.]
    Delta = 8000.
    rtol = 1e-8

    def Avp(v): return problem.Hvp(0, v)

    z = cg_steihaug(problem.vs, Avp, g, g, Delta, rtol, monitor=print)

    assert_allclose(Avp(z), g)

    z = cg_steihaug(problem.vs, Avp, g, g*0, Delta, rtol, monitor=print)

    assert_allclose(Avp(z), g)

    def cg_precond(v, direction):
        if direction == -1:
            return numpy.linalg.inv(C).dot(v)
        else:
            return C.dot(v)

    z = cg_steihaug(problem.vs, Avp, g, g*0, Delta, rtol, monitor=print, C=cg_precond)

    assert_allclose(Avp(z), g)


@pytest.mark.parametrize("precond", [True, False])
def test_tr(precond):
    trcg = TrustRegionCG(maxradius=10., maxiter=100, cg_monitor=print)
    problem = RosenProblem(precond=precond)

    x0 = numpy.zeros(20)
    r = trcg.minimize(problem, x0, monitor=print)
    assert r.converged
    assert_allclose(r.x, 1.0, rtol=1e-4)

@pytest.mark.parametrize("alpha,beta", 
    [
    [1.0, 0.0],
    [0.01, 1.0],
    [1.0, 1.0],
    ]
)
def test_gaussnewton(alpha, beta):
    trcg = TrustRegionCG(maxiter=100, cg_rtol=1e-9, rtol=1e-8, maxradius=80)

    J = numpy.array([ [0, 0,     2,  1],
                      [0,  10,   2,  0],
                      [40, 100,  0,  0],
                      [400, 0,   0,  0]])
    def phi(x): return alpha * x + beta * x ** 2
    def phiprime(x): return alpha + 2 * beta * x

    problem = ChiSquareProblem(J=J, phi=phi, phiprime=phiprime)

    x0 = numpy.zeros(4)
    r = trcg.minimize(problem, x0, monitor=print)
    assert r.converged
    assert_allclose(problem.f(r.x), 0, atol=1e-7)

@pytest.mark.parametrize("alpha,beta", 
    [
    [1.0, 0.0],
    ]
)
def test_gaussnewton_prec(alpha, beta):
    trcg = TrustRegionCG(maxiter=100, cg_rtol=1e-9,
            rtol=1e-8, maxradius=80,
            cg_monitor=print)

    J = numpy.array([ [1e6, 0,     0,  1],
                      [0,  10,   0,  0],
                      [2,  0,  10,  0],
                      [1, 0,   0,  1]])

    # finding the diagonals and use it to precondition
    # the hessian inversion.
    def cg_precond(Avp):
        s = 0
        for v in numpy.eye(len(J)):
            Av = Avp(v)
            vAv = v * Av
            s = s + vAv
        def C(r, direction):
            if direction == -1:
                return r / s
            else:
                return r * s
        return C

    trcg.cg_preconditioner = cg_precond

    def phi(x): return alpha * x + beta * x ** 2
    def phiprime(x): return alpha + 2 * beta * x

    problem = ChiSquareProblem(J=J, phi=phi, phiprime=phiprime)

    x0 = numpy.zeros(len(J))

    r = trcg.minimize(problem, x0, monitor=print)
    assert r.converged
    assert r.nit < 10
    assert_allclose(problem.f(r.x), 0, atol=1e-7)




from __future__ import print_function

import pytest

from abopt.base import Preconditioner

from abopt.linesearch import minpack, backtrace, exact

from abopt.algs.lbfgs import LBFGS
from abopt.algs.lbfgs import inverse_bfgs, direct_bfgs, scalar, inverse_dfp
from abopt.algs.lbfgs import pre_scaled_direct_bfgs, pre_scaled_inverse_dfp
from abopt.algs.lbfgs import post_scaled_direct_bfgs, post_scaled_inverse_dfp

from abopt.testing import RosenProblem, ChiSquareProblem
import numpy
from numpy.testing import assert_allclose

@pytest.mark.parametrize("linesearch, diag_update, rescale_diag",
    [
        (backtrace, post_scaled_direct_bfgs, False), # VAFFast
#        (backtrace, direct_bfgs, True), # VAFGoodHessian many iterations!
#        (backtrace, pre_scaled_direct_bfgs, True), # VAFBadHessian many iterations!
        (minpack, post_scaled_direct_bfgs, False), # VAFFast
        (exact, post_scaled_direct_bfgs, False), # VAFFast
    ]
)
def test_abopt_lbfgs(linesearch, diag_update, rescale_diag):
    lbfgs = LBFGS(linesearch=linesearch,
            diag_update=diag_update,
            rescale_diag=rescale_diag,
            maxiter=1000)

    problem = RosenProblem()

    x0 = numpy.zeros(20)
    r = lbfgs.minimize(problem, x0, monitor=print)
    assert r.converged
    assert_allclose(r.x, 1.0, rtol=1e-4)


def diag_scaling(v, direction):
    if direction == -1:
        return 0.05 * v
    else:
        return 20 * v

precond = Preconditioner(Pvp=diag_scaling, vPp=diag_scaling)
@pytest.mark.parametrize("precond",
[None, precond])
def test_abopt_lbfgs_quad(precond):
    lbfgs = LBFGS(linesearch=backtrace)

    J = numpy.array([ [0, 0,     2,  1],
                      [0,  10,   2,  0],
                      [40, 100,  0,  0],
                      [400, 0,   0,  0]])

    problem = ChiSquareProblem(J=J, precond=precond)

    x0 = numpy.zeros(4)
    r = lbfgs.minimize(problem, x0, monitor=print)
    assert r.converged
    assert_allclose(problem.f(r.x), 0.0, atol=1e-7)



"""
    Vectors.

    define addmul, which allows transporting vectors;
       and dot, which defines the inner product and distance.

    ABOPT sees complex numbers as a tuple of real numbers.
    and thus the vector space is self-dual;
    :math:`x + \lambda g` being well defined.

    `autograd` sees complex numbers as complex numbers,
    thus the vector space is not self-dual; gradient lives in the dual
    space and :math:`x + \lambda g^\dagger` is well defined.

"""
from abopt.base import VectorSpace

class RealVectorSpace(VectorSpace):
    def addmul(self, a, b, c, p=1):
        """ a + b * c ** p, follow the type of b """
        if p is not 1: c = c ** p
        c = b * c
        if a is not 0: c = c + a
        return c

    def dot(self, a, b):
        """ einsum('i,i->', a, b) """
        if hasattr(a, 'dot'):
            return (a * b).sum()
        try:
            return sum(a * b)
        except TypeError:
            return float(a * b)

# helper functions to pack and unpack complex numbers.
def _c2r(a):
    # complex vector space needs numpy
    import numpy
    # in abopt, a scalar almost always means Identity * scalar
    # it must be real; use numpy's broadcast
    if numpy.isscalar(a):
        assert numpy.imag(a) == 0
        return a
    a = numpy.concatenate([numpy.real(a), numpy.imag(a)], axis=0)
    return a


def _r2c(a):
    import numpy
    # in abopt, a scalar almost always means Identity * scalar
    # it must be real; use numpy's broadcast
    if numpy.isscalar(a): 
        assert numpy.imag(a) == 0
        return a
    h = a.shape[0] // 2
    return a[:h] + a[h:] * 1j

class ComplexVectorSpace(VectorSpace):
    def addmul(self, a, b, c, p=1):
        a = _c2r(a)
        b = _c2r(b)
        c = _c2r(c)
        return _r2c(RealVectorSpace.addmul(RealVectorSpace(), a, b, c, p))

    def dot(self, a, b):
        return RealVectorSpace.dot(RealVectorSpace(), _c2r(a), _c2r(b))

real_vector_space = RealVectorSpace()
complex_vector_space = ComplexVectorSpace()



from abopt.base import Proposal

def backtrace(problem, state, z, rate, maxiter, c=1e-5, tau=0.5):
    vs = problem.vs

    addmul = vs.addmul
    dot = vs.dot

    zz = dot(z, z)
    zg = dot(z, state.Pg) / zz ** 0.5

    if zg < 0.0: #1 * state.Pgnorm:
        return None, None

    Px1 = addmul(state.Px, z, -rate)
    prop = Proposal(problem, Px=Px1, z=z).complete_y(state)

    i = 0
    propmin = prop
    ratemin = rate
    while i < maxiter:
        # print('rate', rate, 'y', state.y, 'y1', y1, 'x', state.Px, 'x1', Px1, 'z', z)

        # watch out : do not check convergence ; avoid jumping too far to the other side
        # sufficient descent

        if prop.y < propmin.y:
            propmin = prop
            ratemin = rate
        if prop.y < state.y and abs(prop.y - state.y) >= abs(rate * c * zg):
            # sufficient
            return prop, rate

        rate *= tau
        Px1 = addmul(state.Px, z, -rate)
        prop = Proposal(problem, Px=Px1, z=z).complete_y(state)
        i = i + 1
    return None, None




"""
    Functions of the test suite

    https://en.wikipedia.org/wiki/Test_functions_for_optimization
"""

import autograd
from autograd import numpy
from autograd.numpy import cos, sin, exp, log, pi
from autograd import grad

class FunctionND:
    xatol = 2e-6
    yatol = 1e-7
    def __init__(self, nd):
        self.nd = nd

    def gradient(self, x):
        x = numpy.array(x, dtype='f8')
        return grad(self.function)(x)

    @property
    def start(self):
        return numpy.zeros(self.nd, dtype='f8')

    @property
    def start(self):
        return numpy.zeros(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

class Function2D:
    xatol = 2e-6
    yatol = 1e-7
    def __init__(self, nd):
        assert nd == 2
        self.nd = 2

    def gradient(self, x):
        x = numpy.array(x, dtype='f8')
        return grad(self.function)(x)

    @property
    def start(self):
        return numpy.zeros(self.nd, dtype='f8')

    @property
    def start(self):
        return numpy.zeros(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

class Ackley(Function2D):
    def function(self, x):
        x, y = x
        A = -20 * exp(-0.2 * (0.5 * (x**2 + y**2)) ** 0.5)
        B = - exp(0.5 *(cos(2 * pi * x) + cos(2 * pi * y)))
        C = exp(1) + 20

        return A + B + C

    @property
    def xmin(self):
        return numpy.zeros(self.nd)

    @property
    def start(self):
        return numpy.ones(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return (-5. * numpy.ones(self.nd),
                 5. * numpy.ones(self.nd))

    def gradient(self, x):
        # This function has a nan gradient at 0, 0., not we want..
        x = numpy.array(x, dtype='f8')
        if (x == 0).all():
            return x * 0.0
        return grad(self.function)(x)

class Bukin6(Function2D):
    """ Very hard problem for quasi-newton:
        https://www.sfu.ca/~ssurjano/bukin6.html

    """

    xatol = 8e-2
    yatol = 5e-3

    def function(self, x):
        x, y = x
        return 100 * abs(y - 0.01 * x **2) ** 0.5 + 0.01 * abs(x + 10)

    @property
    def xmin(self):
        return numpy.array([-10, 1.])

    @property
    def start(self):
        return self.xmin + [0.1, 0.1]

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return numpy.array([-15. -5]), numpy.array([-3, 3])

    def gradient(self, x):
        # This function has a nan gradient at 0, 0., not we want..
        x = numpy.array(x, dtype='f8')
        if (x == self.xmin).all():
            return x * 0.0
        return grad(self.function)(x)

class Rastrigin(FunctionND):
    def function(self, x):
        x = numpy.array(x)

        A = 20
        r = A * self.nd
        r = r + numpy.sum(x ** 2 - A * cos(2 * pi * x))
        return r

    @property
    def xmin(self):
        return numpy.zeros(self.nd)

    @property
    def start(self):
        return numpy.ones(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0


    @property
    def domain(self):
        return (-5.12 * numpy.ones(self.nd),
                 5.12 * numpy.ones(self.nd))


class Sphere(FunctionND):
    def function(self, x):
        x = numpy.array(x)
        r = numpy.sum(x ** 2)
        return r

    @property
    def xmin(self):
        return numpy.zeros(self.nd)

    @property
    def start(self):
        return numpy.ones(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return (-5. * numpy.ones(self.nd),
                 5. * numpy.ones(self.nd))

class StyblinskiTang(FunctionND):
    def function(self, x):
        x = numpy.array(x)
        A = numpy.sum(x ** 4)
        B = -16 * numpy.sum(x ** 2)
        C = 5 * numpy.sum(x)
        return (A + B + C) * 0.5

    @property
    def xmin(self):
        return numpy.ones(self.nd) * -2.903534

    @property
    def start(self):
        return self.xmin + 0.1 * numpy.ones(self.nd, dtype='f8')

    @property
    def fmin(self):
        return -39.1661655 * self.nd

    @property
    def domain(self):
        return (-5. * numpy.ones(self.nd),
                 5. * numpy.ones(self.nd))

class Rosenbrock(FunctionND):
    def function(self, x):
        x = numpy.array(x)
        r = 100 * numpy.sum((x[1:] - x[:-1] ** 2)**2) + numpy.sum((1 - x)**2)
        return r

    @property
    def xmin(self):
        return numpy.ones(self.nd)

    @property
    def start(self):
        return numpy.zeros(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return (-5. * numpy.ones(self.nd),
                 5. * numpy.ones(self.nd))

class Beale(Function2D):
    xatol = 1e-5
    yatol = 5e-4 # very slow bottom!
    def function(self, x):
        x, y = x
        A = (1.5 - x + x * y) ** 2
        B = (2.25 -x + x * y**2)**2
        C = (2.625 - x + x * y ** 3) ** 2
        return A + B + C

    @property
    def xmin(self):
        return numpy.array([3, 0.5])

    @property
    def start(self):
        return numpy.ones(self.nd, dtype='f8')

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return (-4.5 * numpy.ones(self.nd),
                 4.5 * numpy.ones(self.nd))

class GoldsteinPrice(Function2D):
    def function(self, x):
        x, y = x

        A = 1 + (x + y + 1)**2 * (19 - 14 * x + 3 * x**2 - 114 * y + 6 * x*y  + 3 * y **2)
        B = 30 + (2  * x - 3 * y) ** 2 * (18 - 32 * x + 12 * x ** 2 + 48 * y - 36 * x * y + 27 * y**2)
        return A * B

    @property
    def xmin(self):
        return numpy.array([0., -1.])

    @property
    def start(self):
        # if too far will run beyond the constraints.
        return self.xmin + [0.1, 0.1]

    @property
    def fmin(self):
        return 3.0

    @property
    def domain(self):
        return (-10 * numpy.ones(self.nd),
                 10 * numpy.ones(self.nd))

class Himmelblau(Function2D):
    def function(self, x):
        x, y = x

        return (x ** 2 + y - 11) ** 2 + (x + y ** 2 - 7) ** 2

    @property
    def xmin(self):
        return numpy.array([3.0, 2.0])

    @property
    def start(self):
        # if too far will run beyond the constraints.
        return numpy.ones(self.nd)

    @property
    def fmin(self):
        return 0.0

    @property
    def domain(self):
        return numpy.array([
                [-5, -5],
                [5, 5]])

    

class McCormick(Function2D):
    xatol = 1e-4
    yatol = 1e-5
    def function(self, x):
        x, y = x

        return sin(x + y) + (x - y)**2 - 1.5 * x + 2.5 * y + 1

    @property
    def xmin(self):
        return numpy.array([-0.54719, -1.54719])

    @property
    def start(self):
        # if too far will run beyond the constraints.
        return numpy.ones(self.nd)

    @property
    def fmin(self):
        return -1.913223

    @property
    def domain(self):
        return numpy.array([
                [-1.5, -3],
                [4, 4]])

def get_all_nd():
    r = []
    d = globals()
    for x in d:
        c = d[x]
        if not isinstance(c, type): continue
        if c is Function2D: continue
        if c is FunctionND: continue
        if(issubclass(c, FunctionND)):
            r.append(c)
    return r

def get_all_2d():
    r = []
    d = globals()
    for x in d:
        c = d[x]
        if not isinstance(c, type): continue
        if c is Function2D: continue
        if c is FunctionND: continue
        if issubclass(c, (Function2D, FunctionND)):
            r.append(c)
    return r



from __future__ import print_function
from abopt.legacy.vmad import VM, microcode, programme, Zero, Tape
from numpy.testing import assert_raises, assert_array_equal, assert_allclose
import numpy

class TestVM(VM):
    @microcode(ain=['x'], aout=['y'])
    def unitary(self, x, y, factor):
        y[...] = x * factor
    @unitary.defvjp
    def _(self, _x, _y, factor):
        _x[...] = _y * factor
    @microcode(ain=['x1', 'x2'], aout=['y'])
    def binary(self, x1, x2, y):
        y[...] = x1 + x2
    @binary.defvjp
    def _(self, _x1, _x2, _y):
        _x1[...] = _y
        _x2[...] = _y

def test_single_compute():
    vm = TestVM()
    code = vm.code()
    code.unitary(x='a', y='b', factor=3.0)
    b = code.compute('b', {'a' : 1.0})
    assert_array_equal(b, 3.0)

def test_single_gradient():
    vm = TestVM()
    code = vm.code()
    code.unitary(x='a', y='b', factor=3.0)
    b, _a = code.compute_with_gradient(['b', '_a'], {'a' : 1.0}, {'_b': 1.0})
    assert_array_equal(b, 3.0)
    assert_array_equal(_a, 3.0)

def test_nested_gradient():
    vm = TestVM()
    code = vm.code()
    code.unitary(x='a', y='b', factor=3.0)
    code.unitary(x='b', y='c', factor=3.0)
    c = code.compute('c', {'a' : 1.0})
    assert_array_equal(c, 9.0)

    _a = code.compute_with_gradient('_a', {'a' : 1.0}, {'_c': 1.0})
    c, _a = code.compute_with_gradient(['c', '_a'], {'a' : 1.0}, {'_c': 1.0})

    assert_array_equal(c, 9.0)
    assert_array_equal(_a, 9.0)

def test_partial_gradient():
    vm = TestVM()
    code = vm.code()
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.unitary(x='a', y='b3', factor=3.0)
    code.unitary(x='a', y='b4', factor=3.0)
    code.binary(x1='b1', x2='b2', y='c1')
    code.binary(x1='b3', x2='b4', y='c2')
    code.binary(x1='c1', x2='c2', y='d')

    d, tape = code.compute('d', {'a' : 1.0}, return_tape=True)
    assert_array_equal(d, 12.0)
    gradient = vm.gradient(tape)

    d, _a = code.compute_with_gradient(['d', '_a'], {'a' : 1.0}, {'_d': 1.0})

    assert_array_equal(d, 12.0)
    assert_array_equal(_a, 12.0)


def test_inplace_gradient():
    vm = TestVM()
    code = vm.code()
    code.unitary(x='a', y='a', factor=3.0)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.binary(x1='b1', x2='b2', y='b1')
    code.unitary(x='b1', y='d', factor=3.0)

    d, tape = code.compute('d', {'a' : 1.0}, return_tape=True)
    assert_array_equal(d, 54.0)
    gradient = vm.gradient(tape)

    d, _a = code.compute_with_gradient(['d', '_a'], {'a' : 1.0}, {'_d': 1.0})

    assert_array_equal(d, 54.0)
    assert_array_equal(_a, 54.0)




from __future__ import print_function
import numpy

from abopt.algs.newton import DirectNewton

from numpy.testing import assert_allclose
from abopt.testing import RosenProblem

import pytest

def diag_scaling(v, direction):
    if direction == -1:
        return 0.05 * v
    else:
        return 20 * v

@pytest.mark.parametrize("precond", [ True, False ])
def test_newton(precond):
    nt = DirectNewton()

    problem = RosenProblem(precond=precond)

    problem.atol = 1e-7 # ymin = 0

    x0 = numpy.zeros(20)
    r = nt.minimize(problem, x0, monitor=print)
    assert_allclose(r.x, 1.0, rtol=1e-4)
    assert r.converged



from abopt.abopt2 import Problem, Preconditioner
from scipy.optimize import rosen, rosen_der, rosen_hess_prod, rosen_hess
import numpy

class ChiSquareProblem(Problem):
    """
        chisquare problem with

        y = |J phi(x) - 1.0|^2
    """
    def __init__(self, J, phi=lambda x:x, phiprime=lambda x: 1, precond=None):
        def f(x): return J.dot(phi(x))
        def vjp(x, v): return v.dot(J) * phiprime(x)
        def jvp(x, v): return J.dot(v * phiprime(x))

        def objective(x):
            y = f(x)
            return numpy.sum((y - 1.0) ** 2)

        def gradient(x):
            y = f(x)
            return vjp(x, y - 1.0) * 2

        def hessian(x, v):
            v = numpy.array(v)
            return vjp(x, jvp(x, v)) * 2

        Problem.__init__(self,
                      objective=objective,
                      gradient=gradient,
                      hessian_vector_product=hessian,
                      precond=precond)

from scipy.linalg import inv
def rosen_inverse_hess(x):
    H = rosen_hess(x)
    return inv(H)

def diag_scaling(v, direction):
    if direction == -1:
        return 0.05 * v
    else:
        return 20 * v

class RosenProblem(Problem):
    """ RosenBrock problem that supports up to hessian inverse. """
    def __init__(self, precond=False):
        if precond == True:
            precond = Preconditioner(Pvp=diag_scaling, vPp=diag_scaling)
        else: 
            precond = None
        Problem.__init__(self, objective=rosen, gradient=rosen_der,
                    hessian_vector_product=rosen_hess_prod,
                    inverse_hessian_vector_product = lambda x, v: rosen_inverse_hess(x).dot(v),
                    precond=precond)



from __future__ import print_function
from abopt.legacy.vmad2 import CodeSegment, Engine, statement, programme, ZERO, logger, Literal
from numpy.testing import assert_raises, assert_array_equal, assert_allclose
from numpy.testing.decorators import skipif
import pytest
xfail = pytest.mark.xfail
import numpy
import logging

try: import graphviz
except ImportError: graphviz = None

logger.setLevel(level=logging.INFO)
class MySubEngine(Engine):
    @statement(ain=['x'], aout=['y'])
    def unitary(engine, x, y, factor):
        y[...] = x * factor

    @unitary.defvjp
    def _(engine, _x, _y, factor):
        _x[...] = _y * factor

    @unitary.defjvp
    def _(engine, x_, y_, factor):
        y_[...] = x_ * factor

class MyEngine(Engine):
    def __init__(self):
        self.subengine = MySubEngine()

    @statement(ain=['x'], aout=['y'])
    def unitary(engine, x, y, factor):
        y[...] = x * factor

    @unitary.defvjp
    def _(engine, _x, _y, factor):
        _x[...] = _y * factor

    @unitary.defjvp
    def _(engine, x_, y_, factor):
        y_[...] = x_ * factor

    @statement(ain=['x1', 'x2'], aout=['y'])
    def binary(engine, x1, x2, y):
        y[...] = x1 + x2

    @binary.defvjp
    def _(engine, _x1, _x2, _y):
        _x1[...] = _y
        _x2[...] = _y

    @binary.defjvp
    def _(engine, x1_, x2_, y_):
        y_[...] = x1_ + x2_

    @programme(ain=['u'], aout=['v'])
    def batch(engine, u, v):
        code = CodeSegment(engine)
        code.unitary(x=u, y=u, factor=1.0)
        code.binary(x1=u, x2=u, y=v)
        return code

    @programme(ain=['u'], aout=['v'])
    def batch_batch(engine, u, v):
        code = CodeSegment(engine)
        code.batch_with_exarg(u=u, v='t', factor=1)
        code.unitary(x='t', y=v, factor=1.0)
        return code

    @programme(ain=['u'], aout=['v'])
    def batch_with_sub(engine, u, v):
        code = CodeSegment(engine.subengine)
        code.unitary(x=u, y=v, factor=2.0)
        return code

    @programme(ain=['u'], aout=['v'])
    def batch_with_exarg(engine, u, v, factor):
        code = CodeSegment(engine)
        code.unitary(x=u, y=u, factor=factor)
        code.binary(x1=u, x2=u, y=v)
        return code

    @programme(ain=['u'], aout=['v'])
    def batch_unused(engine, u, v):
        code = CodeSegment(engine)
        code.unitary(x=Literal(1.0), y=v, factor=1.0)
        return code

def test_compute():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='b', factor=3.0)
    b = code.compute('b', {'a' : 1.0})
    assert_array_equal(b, 3.0)

def test_optimize():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='d', factor=3.0)
    code.unitary(x='a', y='b', factor=3.0)
    code.unitary(x='a', y='c', factor=3.0)

    opt = code.optimize(['b'])
    assert len(opt.nodes) == 1
    b = opt.compute('b', {'a' : 1.0})
    assert_array_equal(b, 3.0)

def test_optimized_execution():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='d', factor=3.0)
    code.unitary(x='a', y='b', factor=3.0)
    code.unitary(x='a', y='c', factor=3.0)

    opt = code.optimize(['b'])
    assert len(opt.nodes) == 1
    
    b, tape = code.compute('b', {'a' : 1.0}, return_tape=True)
    print(tape)

def test_nested_compute():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='b', factor=3.0)
    code.unitary(x='b', y='c', factor=3.0)

    c = code.compute('c', {'a' : 1.0})
    assert_array_equal(c, 9.0)

    c, _a = code.compute_with_gradient(['c', '_a'], {'a' : 1.0}, {'_c': 1.0})

    assert_array_equal(c, 9.0)
    assert_array_equal(_a, 9.0)

def test_tape_gradients():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.unitary(x='a', y='b3', factor=3.0)
    code.unitary(x='a', y='b4', factor=3.0)
    code.binary(x1='b1', x2='b2', y='c1')
    code.binary(x1='b3', x2='b4', y='c2')
    code.binary(x1='c1', x2='c2', y='d')

    d, tape = code.compute('d', {'a' : 1.0}, return_tape=True)
    assert_array_equal(d, 12.0)

    vjp = tape.get_vjp()
    _a  = vjp.compute(['_a'], {'_d': 1.0})
    assert_array_equal(_a, 12.0)

    jvp = tape.get_jvp()
    d_  = jvp.compute(['d_'], {'a_': 1.0})
    assert_array_equal(d_, 12.0)

    d, _a = code.compute_with_gradient(['d', '_a'], {'a' : 1.0}, {'_d': 1.0})

    assert_array_equal(d, 12.0)
    assert_array_equal(_a, 12.0)

    (c1, d), tape = code.compute(['c1', 'd'], {'a' : 1.0}, return_tape=True)

    assert_array_equal(d, 12.0)
    assert_array_equal(c1, 6.0)

    vjp = tape.get_vjp()
    jvp = tape.get_jvp()

    _a = vjp.compute(['_a'], {'_c1': 1.0})
    c1_, d_ = jvp.compute(['c1_', 'd_'], {'a_': 1.0})

    assert_array_equal(c1_, 6.0)
    assert_array_equal(d_, 12.0)

@xfail(reason="jvp from code doesn't work")
def test_jvp():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.unitary(x='a', y='b3', factor=3.0)
    code.unitary(x='a', y='b4', factor=3.0)
    code.binary(x1='b1', x2='b2', y='c1')
    code.binary(x1='b3', x2='b4', y='c2')
    code.binary(x1='c1', x2='c2', y='d')

    jvp = code.get_jvp(init={'a' : 1.0})

    d_ = jvp.compute('d_', {'a_' : 1.0})
    assert_array_equal(d_, 12.0)

    c1_, d_ = jvp.compute(['c1_', 'd_'], {'a_' : 1.0})
    assert_array_equal(d_, 12.0)
    assert_array_equal(c1_, 6.0)

@xfail(reason="jvp from code doesn't work")
def test_jvp_programme():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch(u='a', v='d')
    code.unitary(x='d', y='d', factor=2.0)
    code.batch_with_exarg(u='a', v='e', factor=3.0)

    jvp = code.get_jvp(init={'a' : 1.0})
    d_, e_ = jvp.compute(['d_', 'e_'], {'a_' : 1.0})
    assert_array_equal(d_, 4.0)
    assert_array_equal(e_, 6.0)

@xfail(reason="jvp from code doesn't work")
def test_jvp_programme_nested():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch_batch(u='a', v='d')

    jvp = code.get_jvp(init={'a' : 1.0})

    d_ = jvp.compute('d_', {'a_' : 1.0})
    assert_array_equal(d_, 2.0)

    d, tape = code.compute('d', init={'a' : 1.0}, return_tape=True)
    jvp = tape.get_jvp()

    d_ = jvp.compute('d_', {'a_' : 1.0})
    assert_array_equal(d_, 2.0)

@xfail(reason="jvp from code doesn't work")
def test_jvp_vector():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch(u='a', v='d')
    code.unitary(x='d', y='d', factor=2.0)
    code.batch_with_exarg(u='a', v='e', factor=3.0)

    A = numpy.array

    jvp = code.get_jvp(init={'a' : A([1.0, 1.0])})
    d_, e_ = jvp.compute(['d_', 'e_'], {'a_' : A([1.0, 1.0])})
    assert_array_equal(d_, [4.0, 4.0])
    assert_array_equal(e_, [6.0, 6.0])

def test_inplace():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='a', factor=3.0)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.binary(x1='b1', x2='b2', y='b1')
    code.unitary(x='b1', y='d', factor=3.0)

    d = code.compute('d', {'a' : 1.0})
    assert_array_equal(d, 54.0)

    d, _a = code.compute_with_gradient(['d', '_a'], {'a' : 1.0}, {'_d': 1.0})

    assert_array_equal(d, 54.0)
    assert_array_equal(_a, 54.0)

@skipif(graphviz == None, "graphviz is not properly installed")
def test_to_graph():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch_with_sub(u='a', v='e')
    code.unitary(x='a', y='a', factor=3.0)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.binary(x1='b1', x2='b2', y='b1')
    code.unitary(x='b1', y='d', factor=3.0)
    code.batch(u='b2', v='f')

    d, tape = code.compute(('e', 'a', 'f', 'd'), {'a' : 1.0}, return_tape=True)
    vjp = tape.get_vjp()

    graph1 = code.to_graph()
    graph2 = vjp.to_graph()

#    graph2.render('temp.png', view=True)

def test_zeros():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='a', factor=3.0)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    a, b1, b2, _a = code.compute_with_gradient(['a', 'b1', 'b2', '_a'], {'a' : 1.0}, {'_b1': ZERO, '_b2' : ZERO, '_a' : ZERO})
    assert _a is ZERO

def test_literal():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.binary(x1='a', x2=Literal(2.0), y='a')
    code.batch(u=Literal(2.0), v='d')
    d, a, _a = code.compute_with_gradient(['d', 'a', '_a'], {'a' : 1.0}, {'_a': 1.0, '_d' : 0.0})
    assert_array_equal(a, 3.0)
    assert_array_equal(_a, 1.0)

def test_copy():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch_with_sub(u='a', v='e')
    code.unitary(x='a', y='a', factor=3.0)
    code.unitary(x='a', y='b1', factor=3.0)
    code.unitary(x='a', y='b2', factor=3.0)
    code.binary(x1='b1', x2='b2', y='b1')
    code.unitary(x='b1', y='d', factor=3.0)
    code.batch(u='b2', v='f')

    code2 = code.copy()

def test_programme():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch(u='a', v='d')
    code.batch_with_exarg(u='a', v='e', factor=3.0)
    (d, e), tape = code.compute(('d', 'e'), {'a' : 1.0}, return_tape=True)
    assert_array_equal(d, 2.0)
    assert_array_equal(e, 6.0)
    e, d, _a = code.compute_with_gradient(['e', 'd', '_a'], {'a' : 1.0}, {'_d': 1.0, '_e' : 0.0})
    assert_array_equal(d, 2.0)
    assert_array_equal(e, 6.0)
    assert_array_equal(_a, 2.0)
    e, d, _a = code.compute_with_gradient(['e', 'd', '_a'], {'a' : 1.0}, {'_d': 0.0, '_e' : 1.0})
    assert_array_equal(d, 2.0)
    assert_array_equal(e, 6.0)
    assert_array_equal(_a, 6.0)

def test_programme_unused():
    engine = MyEngine()
    code = CodeSegment(engine)
    code.batch_unused(u='a', v='d')
    d, _a = code.compute_with_gradient(['d', '_a'], {'a' : 1.0}, {'_d': 1.0})
    assert_array_equal(d, 1.0)
    assert_array_equal(_a, 0.0)




from abopt.base import Proposal

def exact(problem, state, z, rate, maxiter, c=0.5):
    vs = problem.vs
    addmul = vs.addmul
    dot = vs.dot

    znorm = dot(z, z) ** 0.5

    from scipy.optimize import minimize_scalar

    Px1 = addmul(state.Px, z, - rate)
    prop = Proposal(problem, Px=Px1, z=z).complete_y(state)

    best = [prop, 1.0]

    def func(tau):
        if tau == 0: return state.y

        Px1 = addmul(state.Px, z, -tau * rate)
        prop = Proposal(problem, Px=Px1, z=z).complete_y(state)

        if prop.y < best[0].y:
            best[0] = prop
            best[1] = tau

        if prop.y < state.y and abs(prop.y - state.y) >= abs(c * znorm):
            raise StopIteration

        return prop.y

    try:
        r = minimize_scalar(func, (0, 1), bounds=(0, 1), method='brent', options={'maxiter':10}, )

        if not r.success or r.fun >= state.y:
            raise StopIteration

        Px1 = addmul(state.Px, z, -r.x * rate)
        return Proposal(problem, Px=Px1, y=r.fun, z=z), r.x * rate

    except StopIteration as e:
        return best[0], best[1] * rate




from __future__ import print_function

from abopt.abopt2 import GradientDescent, LBFGS, Preconditioner, minimize

from abopt.linesearch import minpack, backtrace, exact
from abopt.vectorspace import real_vector_space, complex_vector_space

from numpy.testing import assert_raises, assert_allclose

from scipy.optimize import rosen, rosen_der
import numpy

def quad(x):
    return (x[0] - .5)** 2 + (x[1] - .5) ** 2
def quad_der(x):
    return (x - 0.5) * 2

def crosen(x):
    x = numpy.atleast_1d(x)
    v = numpy.concatenate([x.real, x.imag], axis=0)
    return rosen(v)

def crosen_der(x):
    x = numpy.atleast_1d(x)
    s = x.shape[0]
    v = numpy.concatenate([x.real, x.imag], axis=0)
    v = rosen_der(v)
    r = v[:s] + v[s:] * 1j
    return r

x0 = numpy.zeros(2)


def test_abopt_gd_backtrace():
    gd = GradientDescent(linesearch=backtrace)

    s = minimize(gd, quad, quad_der, x0, monitor=print)
    print(s)
    assert s.converged
    assert_allclose(s.x, 0.5, rtol=1e-4)

def test_abopt_gd_minpack():
    gd = GradientDescent(linesearch=minpack)
    
    s = minimize(gd, quad, quad_der, x0)
    print(s)
    assert s.converged
    assert_allclose(s.x, 0.5, rtol=1e-4)

def test_abopt_gd_exact():
    gd = GradientDescent(linesearch=exact, maxiter=10)
    
    s = minimize(gd, quad, quad_der, x0, monitor=print)
    assert s.converged
    assert_allclose(s.x, 0.5, rtol=1e-4)

def test_abopt_gd_complex():
    gd = GradientDescent(linesearch=exact, maxiter=10)

    X = []
    Y = []
    def monitor_r(state):
        X.append(state.x[0] + state.x[1] * 1j)

    def monitor_c(state):
        Y.append(state.x[0])

    s = minimize(gd, rosen, rosen_der, numpy.array([0., 0.]), monitor=monitor_r, vs=real_vector_space)
    s = minimize(gd, crosen, crosen_der, numpy.array([0. + 0.j]), monitor=monitor_c, vs=complex_vector_space)

    assert_allclose(X, Y, rtol=1e-4)

def test_state():
    from abopt.abopt2 import State
    s = State()
    print()
    print(s.format(header=True))
    print(s.format())
    print(s.format(columns=['nit', 'na']))



def default_addmul(a, b, s):
    if s is 0:
        return 1.0 * a # always a new instance is created
    if a is 0:
        return b * s
    return a + b * s

def default_dot(a, b):
    if hasattr(a, 'dot'):
        return a.dot(b)
    try:
        return sum(a * b)
    except TypeError:
        return float(a * b)

class parameter(object):
    """ The parameter decorator declares a parameter to
        the optimizer object.

        This class follows the descriptor pattern in Python.

        The return value of the accessor is used to set the attribute.
    """
    def __init__(self, convert):
        self.name = convert.__name__
        self.default = convert.__defaults__[0]
        self.convert = convert
        self.__doc__ = convert.__doc__

    def __get__(self, instance, owner):
        if isinstance(instance, Optimizer):
            return instance.config.get(self.name, self.default)
        else:
            return self

    def __set__(self, instance, value):
        instance.config[self.name] = self.convert(value)

class ConvergenceStatus(BaseException): pass 
class TooManySteps(ConvergenceStatus): pass
class BadDirection(ConvergenceStatus): pass
class Converged(ConvergenceStatus): pass

class Optimizer(object):
    @parameter
    def tol(value=1e-6):
        """Relative tolerance of change in objective. Terminate if dy < tol * y + atol"""
        assert value >= 0
        return value

    @parameter
    def atol(value=0):
        """Absolute tolerance of change objective. Terminate if dy < tol * y + atol """
        assert value >= 0
        return value

    @parameter
    def ymin(value=None):
        """ending objective. Terminate if y < ymin """
        return value

    @parameter
    def gtol(value=1e-6):
        """Absolute tolerance of gradient. Terminate if gnorm < gtol """
        assert value >= 0
        return value

    @parameter
    def maxsteps(value=1000):
        """Maximium  number of iterations"""
        return int(value)

    @parameter
    def csteps(value=3):
        """ number of iterations dy < threshold before confirming convergence """
        return int(value)

    @parameter
    def minsteps(value=10):
        """ minimum number of steps """
        return int(value)

    @parameter
    def local_linesearch(value=True):
        """ whether the blind search tries to use local minima """
        return bool(value)

    @parameter
    def use_linesearch(value=True):
        """ whether use line search """
        return bool(value)

    def copy(self, a):
        return self.addmul(a, a, 0)

    def mul(self, a, s):
        return self.addmul(0, a, s)

    def __setattr__(self, key, value):
        # only allow setting parameters
        if hasattr(type(self), key) and isinstance(getattr(type(self), key), parameter):
            return object.__setattr__(self, key, value)
        else:
            raise AttributeError("Setting attribute %s on an Optimizer of type %s is not supported" % (key, type(self)))

    def __init__(self,
                 addmul=default_addmul,
                 dot=default_dot,
                 ):
        """
            Parameters
            ----------
            addmul : callable
                addmul(a, b, s) returns a + b * s as a new vector from vector a, b and a Python scalar s.
                when s is 0 (not 0.0), it returns a copy of a, serving as a constructor of new vectors.
                when a is 0, it returns b * s. The default is simply `a + b * s` with optimizations
                for zeros.

            dot : callable
                dot(a, b) returns the inner product of vectors a and b as a Python scalar; the default
                works for most cases by first looking for a `dot` method, then fallback to `sum` of
                the `*` operator.
        """
        # FIXME: use check the function signature is correct.
        d = self.__dict__
        d['dot'] = dot
        d['addmul'] = addmul
        d['config'] = {}

    def minimize(self, objective, gradient, x0, monitor=None):
        raise NotImplementedError

    def get_thresh(self, y1, y2):
        valmax = max(abs(y1), abs(y2))
        thresh = self.tol * max(valmax, 1.0) + self.atol
        return thresh

    def converged(self, state, y1):
        return abs(y1 - state.y) < self.get_thresh(state.y, y1) and state.y >= y1

    def backtrace_optimal(self, objective, state, z, zg, rate):
        """ This finds the minimal along the proposal direction. """
        from scipy.optimize import minimize_scalar
        def func(tau):
            if tau == 0: return state.y
            x1 = self.addmul(state.x, z, -tau * rate)
            state.fev = state.fev + 1
            return objective(x1)

        r = minimize_scalar(func, (0, 1), bounds=(0, 1), method='bounded', options={'maxiter':10}, )

        x1 = self.addmul(state.x, z, -r.x * rate)
        return x1, r.fun

    def backtrace(self, objective, state, z, zg, rate):

        # doing only backtracking line search
        # FIXME: implement more-thuente
        tau = 0.5
        c = 1e-5
        x1 = self.addmul(state.x, z, -rate)
        y1 = objective(x1)
        state.fev = state.fev + 1
        i = 0
        while i < 10:
            valmax = max(abs(y1), abs(state.y))
            thresh = self.tol * max(valmax, 1.0) + self.atol

            #print(rate, state.y, y1, state.x, x1, z)
            #if self.converged(state, y1): break

            # sufficient descent
            if y1 < state.y and abs(y1 - state.y) >= abs(rate * c * zg):
                break

            rate *= tau
            x1 = self.addmul(state.x, z, -rate)
            y1 = objective(x1)
            #print('new point ', x1, y1, state.x, state.y)
            state.fev = state.fev + 1
            i = i + 1
        return x1, y1

    def linesearch(self, objective, gradient, state, z, zg, rate):
        # This tries to go a bigger step until the direction of
        # gradient changes or the objective increases.
        # Useful when there is no hessian
        # e.g. with gradient descent but really just a hack -- it may hop around
        # to a different local stationary point.
        # need a better algorithm than this..
        if not self.use_linesearch:
            return self.backtrace(objective, state, z, zg, rate)
        tau = 2.0
        x1 = self.addmul(state.x, z, -rate)
        y1 = objective(x1)
        state.fev = state.fev + 1
        dy = state.y - y1
        while True:
            rate *= tau
            x1 = self.addmul(state.x, z, -rate)
            y1 = objective(x1)
            state.fev = state.fev + 1

            if self.local_linesearch:
                g1 = gradient(x1)
                state.gev = state.gev + 1

                # change direction
                if self.dot(g1, state.g) / state.gnorm < 0.01:
                    break

            # worse descent
            if state.y - y1 < dy:
                break
            dy = state.y - y1


        return self.backtrace(objective, state, z, zg, rate)

class BaseState(object):
    #__slots__ = ['it', 'fev', 'gev', 'dy', 'statue']

    @property
    def x(self): return self._x
    @property
    def xnorm(self): return self._xnorm

    @x.setter
    def x(self, value):
        self._x = value
        self._xnorm = self.optimizer.dot(value, value) ** 0.5

    @property
    def g(self): return self._g
    @property
    def gnorm(self): return self._gnorm

    @g.setter
    def g(self, value):
        self._g = value
        self._gnorm = self.optimizer.dot(value, value) ** 0.5

    def __getitem__(self, name):
        return getattr(self, name)

    def __init__(self, optimizer, objective, gradient, x0):
        self.__dict__['optimizer'] = optimizer

        if isinstance(x0, type(self)):
            self.__dict__.update(x0.__dict__)
        else:
            self.x = optimizer.copy(x0)
            self.dy = None

            self.fev, self.gev = 0, 0
            self.it = 0

        self.status = None
        self.y = objective(self.x)
        self.g = gradient(self.x)
        self.fev = self.fev + 1
        self.gev = self.gev + 1


    def __str__(self):
        d = {}
        d['it'] = self['it']
        d['y'] = self['y']
        d['xnorm'] = self['xnorm']
        d['gnorm'] = self['gnorm']
        d['fev'] = self['fev']
        d['gev'] = self['gev']
        d['dy'] = self['dy']
        if self['dy'] is None:
            d['dy'] = 'None'
        else:
            d['dy'] = '%g' % self['dy']
        return "Iteration %(it)d: y = %(y)g dy = %(dy)s fev = %(fev)d gev = %(gev)d gnorm = %(gnorm)g xnorm = %(xnorm)g" % d

class GradientDescent(Optimizer):
    """ GradientDescent ignores minsteps, csteps, tol and atol. It always run for maxsteps
         -- Since there is no linear search we can never know about the convergence.
    """
    @parameter
    def gamma(value=1e-3):
        """descent rate parameter"""
        assert value > 0
        return value

    class State(BaseState):
        def __init__(self, optimizer, objective, gradient, x0):
            BaseState.__init__(self, optimizer, objective, gradient, x0)

    def minimize(self, objective, gradient, x0, monitor=None):
        state = self.State(self, objective, gradient, x0)

        while state.it < self.maxsteps:
            if monitor:
                monitor(state)

            if state.gnorm < self.gtol: break
            if self.ymin is not None and state.y < self.ymin : break
            # move to the next point
            x1 = self.addmul(state.x, state.g, -self.gamma)
            y1 = objective(x1)
            state.fev = state.fev + 1

            state.g = gradient(state.x)
            state.gev = state.gev + 1


            state.dy = abs(y1 - state.y)
            state.x = x1
            state.y = y1
            state.it = state.it + 1

        return state

class LBFGS(Optimizer):
    @parameter
    def m(value=10):
        """number of vectors for approximating Hessian"""
        return int(value)

    class State(BaseState):
        def __init__(self, optimizer, objective, gradient, x0):
            BaseState.__init__(self, optimizer, objective, gradient, x0)
            self.rho = []
            self.S = []
            self.Y = []
            self.H0k = 1.0
            self.converged_it = 0
            self.z = self.g
            self.use_steepest_descent = False

        def __str__(self):
            return BaseState.__str__(self) + ' GD: %s' % self.use_steepest_descent

    def one(self, objective, gradient, state):
        q = self.copy(state.g)
        alpha = []
        for i in range(len(state.S)):
            dotproduct = self.dot(state.S[i], q)
            alpha.append(state.rho[i] * dotproduct)
            q = self.addmul(q, state.Y[i], -alpha[i])

        z = self.mul(q, state.H0k)

        for i in reversed(list(range(len(state.S)))):
            dotproduct = self.dot(state.Y[i], z)
            beta = state.rho[i] * dotproduct
            z = self.addmul(z, state.S[i], alpha[i] - beta)

        state.use_steepest_descent = False

        znorm = self.dot(z, z) ** 0.5
        zg = 0.0 if znorm == 0 else self.dot(z, state.g) / znorm
        zg_grannorm = 0.0 if state.gnorm == 0 else zg / state.gnorm

        if zg_grannorm < 0.01:
            # L-BFGS gave a bad direction.
            z = self.copy(state.g)
            zg = 1.0
            state.use_steepest_descent = True
        """
        print('--------')
        print(alpha)
        print([1 / tmp for tmp in state.rho])
        print('q', q)
        print('z', z)
        print(state.H0k)
        print(state.Y)
        print(state.S)
        """
        state.z = z
        if state.it == 0 or state.use_steepest_descent:
            rate = 1.0 / state.gnorm
            #x1, y1 = self.linesearch(objective, gradient, state, z, zg, rate)
            x1, y1 = self.backtrace(objective, state, z, zg, rate)
        else: 
            rate = 1.0
            x1, y1 = self.backtrace(objective, state, z, zg, rate)

        g1 = gradient(x1)
        state.gev = state.gev + 1

        if self.converged(state, y1):
            state.converged_it = state.converged_it + 1
        else:
            state.converged_it = 0
            # hessian update

            # XXX: shall we do this when use_steepest_descent is true?
            state.S.insert(0, self.addmul(x1, state.x, -1))
            state.Y.insert(0, self.addmul(g1, state.g, -1))
            ys = self.dot(state.S[0], state.Y[0])
            yy = self.dot(state.Y[0], state.Y[0])

            if ys == 0.0:
                state.status = BadDirection("LBFGS didn't move for some reason ys is 0, QUITTING")
            if yy == 0.0:
                state.status = BadDirection("LBFGS didn't move for some reason yy is 0, QUITTING")

            state.rho.insert(0, 1.0 / ys)

            if len(state.S) > self.m:
                del state.S[-1]
                del state.Y[-1]
                del state.rho[-1]

            state.H0k = ys / yy

        state.dy = abs(y1 - state.y)
        state.x = x1
        state.y = y1
        state.g = g1

        state.it = state.it + 1


    def minimize(self, objective, gradient, x0, monitor=None):
        state = self.State(self, objective, gradient, x0)

        if monitor: monitor(state)

        while True:
            self.one(objective, gradient, state)

            if monitor:
                monitor(state)

            if state.status is not None:
                break

            if state.converged_it >= self.csteps and state.it >= self.minsteps:
                state.status = Converged("YES: Tolerance achieved")
                break

            if state.gnorm < self.gtol:
                state.status = Converged("YES: Gradient tolerance achieved")

            if self.ymin is not None and state.y < self.ymin : 
                state.status = Converged("YES: Objective below threshold.")

            if state.it > self.maxsteps:
                state.status = TooManySteps("maximum number of iterations reached. QUITTING.")

        return state

class SLBFGS(LBFGS):
    """ Stochastic LBFGS """
    @parameter
    def oracle(value=None):
        """Function to generates a random displacement in the parameter space """
        return value

    @parameter
    def N0(value=10):
        return int(value)

    @parameter
    def gthresh(value=1.0):
        return float(value)

    @parameter
    def noiseamp(value=0.1):
        return float(value)

    class State(LBFGS.State):
        def __init__(self, optimizer, objective, gradient, x0):
            LBFGS.State.__init__(self, optimizer, objective, gradient, x0)
            self.it_noise = - optimizer.N0 - 1
            self.x_noise = None
            self.y_noise = None
            self.g_noise = None
            self.noise = None

    def one(self, objective, gradient, state):
        state.noise = None

        if state.it - state.it_noise > self.N0 and state.gnorm < self.gthresh:
            state.x_noise = state.x
            state.it_noise = state.it
            state.g_noise = state.g
            state.y_noise = state.y
            noise = self.oracle(state)
            state.noise = noise
            state.x = self.addmul(state.x_noise, state.noise, self.gthresh * self.noiseamp)
            print('hop', state.x, state.x_noise, state.noise, self.gthresh, self.noiseamp)
            state.y = objective(state.x)
            state.g = gradient(state.x)
            state.fev = state.fev + 1
            state.gev = state.gev + 1
            state.it = state.it + 1
        elif state.it - state.it_noise == self.N0 and state.y - state.y_noise > - self.get_thresh(state.y, state.y_noise):
            state.x = state.x_noise
            state.y = state.y_noise
            state.g = state.g_noise
            #state.status = Converged("Perturbed Gradient Descent converged")
            state.it = state.it + 1
        else:
            state.status = LBFGS.one(self, objective, gradient, state)




from abopt.base import Optimizer
from abopt.linesearch import backtrace

class GradientDescent(Optimizer):

    optimizer_defaults = {
        'maxiter' : 100000,
        'conviter' : 1,
        'linesearch' : backtrace,
        'linesearchiter' : 100,
    }

    def start(self, problem, state, x0):
        prop = Optimizer.start(self, problem, state, x0)
        prop.rate = 1.0
        return prop

    def move(self, problem, state, prop):
        state.rate = prop.rate
        Optimizer.move(self, problem, state, prop)

    def single_iteration(self, problem, state):
        mul = problem.vs.mul

        z = mul(state.Pg, 1 / state.Pgnorm)

        prop, r1 = self.linesearch(problem, state, z, state.rate * 2, maxiter=self.linesearchiter)

        prop.rate = r1
        return prop



"""

    Data model

    A ``Problem`` is defined on a ``VectorSpace``; the problem is to minimize a function to a given tolerance.
    A ``Problem`` consists of a differentiable function, up to second order. `Gradient` and `HessianVectorProduct`
    A ``Problem`` can be `minimize`d by an ``Optimizer``, yielding a sequence of `State`s.
    A ``Problem`` is ``Preconditioned``, the ``Optimizer`` only operates on preconditioned variables.
    An ``Optimizer`` implements a minimization policy (algorithm)

    Problem parameters and Optimizer parameters
    -------------------------------------------
    Problem parameters are related to the accuracy, atol, rtol, etc.
    Optimizer parameters only controls the behavior of the optimizer; maxiter, etc.

    An easy way to see this is that if we redefine the vector variable by a factor of 10,
    if a parameters shall be adjusted, then it belongs to the problem;
    if it shall not be adjusted, then it belongs to the optimizer.
"""

class ContinueIteration(str): pass
class ConvergedIteration(str): pass
class FailedIteration(str): pass

import time

class State(object):
    def __init__(self):
        self.nit = 0
        self.fev = 0
        self.gev = 0
        self.hev = 0
        self.y = None
        self.dy = None
        self.dxnorm = None
        self.xnorm = None
        self.gnorm = None
        self.Pxnorm = None
        self.Pgnorm = None
        self.dxnorm = None
        self.assessment = None
        self.conviter = 0
        self.converged = False
        self.message = ""
        self.y_ = []
        self.z = None
        self.Pg = None
        self.timestamp = time.time()
        self.wallclock = 0

        self.default_format = dict(
        [
            ('wallclock', '[ %08.4f ]',),
            ('nit', '%04d',),
            ('fev', '%04d',),
            ('gev', '%04d',),
            ('hev', '%04d',),
            ('y', '% 13.6e'),
            ('dy', '% 13.6e'),
            ('xnorm', '% 11.4e'),
            ('gnorm', '% 11.4e'),
            ('theta', '% 4.2f'),
            ('radius', '% 9.2e'),
            ('B', '%10s'),
            ('rate', '% 8.2f'),
            ('rho', '% 5.2f'),
            ('conviter', '%04d'),
            ('converged', '% 6s'),
            ('message', '% 20s'),
            ('assessment', '% 20s')
        ])


    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __contains__(self, key):
        return hasattr(self, key)

    def __repr__(self):
        return self.format()

    def format(self, columns=None, header=False, sp='|'):
        """ format a state object.

            Parameters
            ----------
            columns : list of string or tuples.
                for each item,
                if tuple, the format of the column. if string, use a default format
                for the column.
            header : bool
                format the column headers if True
        """

        dd = dict(self.default_format)

        if columns is None:
            columns = self.default_format

        c2 = []
        for item in columns:
            if not isinstance(item, tuple):
                item = (item, dd.get(item, '%10s'))
            c2.append((item[0], item[1]))

        keys = { key : fmt for key, fmt in c2}

        def get_width(key):
            fmt = keys[key]
            width = fmt[1:-1].split('.')[0]
            while len(width) and width[0] not in '0123456789':
                width = width[1:]
            width = int(width.strip())
            if width < len(key):
                width = len(key)
            return width

        def get_strfmt(key, strict):
            if strict:
                return ('%%%d.%ds' % (get_width(key), get_width(key)))
            else:
                return ('%%%ds' % (get_width(key)))

        if header:
            return (sp.join(get_strfmt(key, True) % key for key, fmt in c2))

        else:
            def fmt_field(key):
                if key not in self:
                    s = "N/A"
                else:
                    try:
                        s = keys[key] % self[key]
                    except TypeError:
                        s = str(self[key])

                return get_strfmt(key, False) % s

            return (sp.join(fmt_field(key)  for key, fmt in c2))

class Proposal(object):
    def __init__(self, problem, y=None, x=None, Px=None, g=None, Pg=None, z=None):
        """ A proposal is a collection of variable and gradients.

            We will generate the variables if they are not provided.

        """
        if x is None and Px is not None:
            x = problem.Px2x(Px)

        if Pg is None and g is not None:
            Pg = problem.g2Pg(g)

        self.y = y
        self.x = x
        self.Px = Px
        self.g = g
        self.Pg = Pg
        self.z = z
        self.problem = problem
        self.message = "normal"

    def complete(self, state):
        dot = self.problem.vs.dot
        addmul = self.problem.vs.addmul
        self.znorm = dot(self.z, self.z) ** 0.5
        self.xnorm = dot(self.x, self.x) ** 0.5
        self.Pxnorm = dot(self.Px, self.Px) ** 0.5
        self.complete_y(state)
        self.complete_g(state)

        self.dy = self.y - state.y
        dx = addmul(self.x, state.x, -1)
        self.dxnorm = dot(dx, dx) ** 0.5
        if state.Pgnorm == 0:
            self.theta = 1
        else:
            self.theta = dot(self.z, state.Pg) / (self.znorm * state.Pgnorm)
        return self

    def complete_y(self, state):
        problem = self.problem

        if self.y is None:
            self.y = problem.f(self.x)
            state.fev = state.fev + 1
        return self

    def complete_g(self, state):
        dot = self.problem.vs.dot
        problem = self.problem

        # fill missing values in prop
        if self.g is None:
            self.g = problem.g(self.x)
            state.gev = state.gev + 1

        if self.Pg is None:
            self.Pg = problem.g2Pg(self.g)

        self.Pgnorm = dot(self.Pg, self.Pg) ** 0.5
        self.gnorm = dot(self.g, self.g) ** 0.5

        return self

class InitialProposal(Proposal):
    def complete(self, state):
        dot = self.problem.vs.dot
        addmul = self.problem.vs.addmul
        self.xnorm = dot(self.x, self.x) ** 0.5
        self.Pxnorm = dot(self.Px, self.Px) ** 0.5
        self.complete_y(state)
        self.complete_g(state)

        self.dy = None
        self.dxnorm = None
        self.znorm = None
        self.theta = None
        return self

class Preconditioner(object):
    """ A preconditioner has four functions, corresponding
        to left dot and right dot of P and Q on a vector.

        P = [Q ^{-1}]^T,

        P and Q are Given by the following coordinate transformations
        from x to x~.

            x~ i  = P_ij x_j -> Pvp(x)

            x_j = Q_ij x~_i -> vQp(x~)

        The gradient transformation is from chain rule,

            g~_i = g_j Q_ij -> Qvp(g)
            g_j = g~_i P_ij -> vPp(g~)

        Hessian vector product and inverse Hessian vector product,

            H~_ij v_j = Q_ia Q_jb v_j H_ab -> Qvp(Hvp(vQp(v)))
            h~_ij v_j = P_ai P_bj v_j h_ab -> vPp(hvp(Pvp(v)))

        The input functions are

            Pvp(v, direction),
            vPp(v, direction);
    """
    def __init__(self, Pvp, vPp):
        self.Pvp = Pvp
        self.vPp = vPp

class Problem(object):
    """ Defines a problem.

    """
    def __init__(self, objective, gradient,
        hessian_vector_product=None,
        inverse_hessian_vector_product=None,
        vs=None,
        atol=0,
        rtol=1e-7,
        xtol=1e-7,
        gtol=1e-8,
        precond=None,
        ):
        if precond is None:
            precond = Preconditioner(lambda x, direction:x, lambda x, direction:x)

        if vs is None:
            from .vectorspace import real_vector_space
            vs = real_vector_space

        if not isinstance(vs, VectorSpace):
            raise TypeError("expecting a VectorSpace object for vs, got type(vs) = %s", repr(type(vs)))

        if not isinstance(precond, Preconditioner):
            raise TypeError("expecting a VPreconditioner object for precond, got type(vs) = %s", repr(type(precond)))

        self._precond = precond
        self.vs = vs

        self._objective = objective
        self._gradient = gradient
        self._hessian_vector_product = hessian_vector_product
        self._inverse_hessian_vector_product = inverse_hessian_vector_product
        self.atol = atol
        self.rtol = rtol
        self.xtol = xtol
        self.gtol = gtol

    def Px2x(self, Px):
        return self._precond.vPp(Px, direction=-1)

    def x2Px(self, x):
        return self._precond.Pvp(x, direction=1)

    def g2Pg(self, g):
        return self._precond.Pvp(g, direction=-1)

    def Pg2g(self, Pg):
        return self._precond.vPp(Pg, direction=1)

    def check_preconditioner(self, x0):
        vs = self.vs

        d = vs.addmul(self.Px2x(self.x2Px(x0)), x0, -1)
        if vs.dot(d, d) > 1e-6 * vs.dot(x0, x0):
            raise ValueError("Preconditioner's vQp and Pvp are not inverses.")

        d = vs.addmul(self.Pg2g(self.g2Pg(x0)), x0, -1)
        if vs.dot(d, d) > 1e-6 * vs.dot(x0, x0):
            raise ValueError("Preconditioner's vPp and Qvp are not inverses.")


    def f(self, x):
        return self._objective(x)

    def g(self, x):
        """ This returns the gradient for the original variable"""
        g = self._gradient(x)
        return g

    def Hvp(self, x, v):
        """ This returns the raw hessian product H_x v
            uppercase H means Hessian, not Hessian inverse.

            v is not preconditioned.
            x is not preconditioned.

            result is not preconditioned, and act like x.
        """
        if self._hessian_vector_product is None:
            raise ValueError("hessian vector product is not defined")
        return self._hessian_vector_product(x, v)


    def PHvp(self, x, v):
        """ This returns the preconditioned hessian times v

            uppercase H means Hessian, not Hessian inverse.

            v is usually preconditioned (irrelevant)
            x is always not preconditioned.

            result is preconditioned, and act like Px.

            ~H_x v = Q [H_x (v Q^T)]
        """
        if self._hessian_vector_product is None:
            raise ValueError("hessian vector product is not defined")
        vQ = self._precond.vPp(v, direction=-1)
        return self._precond.Pvp(self._hessian_vector_product(x, vQ), direction=-1)

    def Phvp(self, x, v):
        """ This returns the preconditioned inverse hessian times v

            lowercase h means inverse of Hessian

            v is usually preconditioned (irrelevant)
            x is always not preconditioned.

            result is preconditioned, and act like Px.

            ~h_x v = [h_x (P v)] P^T
        """
        if self._inverse_hessian_vector_product is None:
            raise ValueError("inverse_hessian vector product is not defined")
        Pv = self._precond.Pvp(v, direction=1)
        return self._precond.vPp(self._inverse_hessian_vector_product(x, Pv), direction=1)

    def get_ytol(self, y):
        thresh = self.rtol * abs(y) + self.atol
        return thresh

    def check_convergence(self, y0, y1):
        # This condition may need to be removed.
        # some optimizers do not need the objectve to have been strictly decreasing
        # for convergence.
        if y1 > y0 :
            return False

        valmax = max(abs(y0), abs(y1))

        thresh = self.get_ytol(valmax)

        if abs(y0 - y1) < thresh: return True

        return False


class Optimizer(object):
    optimizer_defaults = {}

    def __init__(self, **kwargs):
        # this updates the attributes
        self.__dict__.update(type(self).optimizer_defaults)
        self.__dict__.update(kwargs)

    def terminated(self, problem, state):
        # check for maxiter first to overrride
        # continuing due to ConvergedItreration.

        if state.nit > self.maxiter: return True

        if isinstance(state.assessment, (ConvergedIteration, FailedIteration)):
            if state.Pgnorm == 0:
                return True
            if state.conviter >= self.conviter:
                return True
            else:
                return False

        if state.dy is None: return False

        return False

    def move(self, problem, state, prop):

        state.message = prop.message

        state.y_.append(prop.y)

        if len(state.y_) > 2: # only store a short history
            del state.y_[0]

        state.y = prop.y
        state.dy = prop.dy

        state.x = prop.x
        state.g = prop.g
        state.z = prop.z
        state.theta = prop.theta
        state.Px = prop.Px
        state.Pg = prop.Pg

        state.xnorm = prop.xnorm
        state.gnorm = prop.gnorm
        state.Pxnorm = prop.Pxnorm
        state.Pgnorm = prop.Pgnorm
        state.dxnorm = prop.dxnorm
        state.wallclock = time.time() - state.timestamp
        state.timestamp = timestamp = time.time()

    def assess(self, problem, state, prop):
        if prop is None:
            return FailedIteration("no proposal is made")

        prop = prop.complete(state)

        if prop.gnorm <= problem.gtol:
            return ConvergedIteration("Gradient is sufficiently small")

        if prop.dxnorm <= problem.xtol:
            return ConvergedIteration("Solution stopped moving")

        if problem.check_convergence(state.y, prop.y):
            return ConvergedIteration("Objective stopped improving")

        return ContinueIteration("continue iteration")

    def single_iteration(self, problem, state):
        # it shall return a Proposal object
        raise NotImplementedError
        # here is an example that doesn't yield a new solution
        return Proposal(Px=state.Px)

    def start(self, problem, state, x0):
        # make a full initial proposal with x and g
        Px0 = problem.x2Px(x0) # the only place we convert from x to Px
        prop = InitialProposal(problem, x=x0, Px=Px0).complete(state)
        return prop

    def _minimize(optimizer, problem, state, monitor=None):

        first_iteration = True
        while not optimizer.terminated(problem, state):
            if isinstance(state.assessment, (ConvergedIteration, FailedIteration)) \
              or first_iteration:
                prop = optimizer.start(problem, state, state['x'])

                optimizer.move(problem, state, prop)

                if monitor is not None:
                    monitor(state)

            first_iteration = False

            prop = optimizer.single_iteration(problem, state)

            # assessment must be before the move, for it needs to see dy
            assessment = optimizer.assess(problem, state, prop)

            state.assessment = assessment

            if isinstance(assessment, (ContinueIteration, ConvergedIteration)):
                optimizer.move(problem, state, prop)
                state.nit = state.nit + 1

            if isinstance(assessment, (ConvergedIteration, FailedIteration)):
                state.conviter = state.conviter + 1
                # converged or failed -- restart cleanly
                state.converged = True

            if isinstance(assessment, ContinueIteration):
                state.conviter = 0
                state.converged = False

            if monitor is not None:
                monitor(state)

        return state

    def minimize(optimizer, problem, x0, monitor=None, **state_args):
        """ minimize a problem starting from state x0

            Parameters
            ----------
            problem : Problem
                the problem object, which defines the objective function
                and the vector space of the parameters.

            x0 : object or a State
                the initial state.
                if a object is given, a state is constructed by setting state_args
                positioning the parameters at x0.
                if a State object is given it will be used and updated.

            monitor: function(state)
                a function that gets called on each iteration.
                if state.dy is None then it is the first time monitor is called.

            Returns
            -------
            state : a State object of the final minimization result.
            state.converged : if the solution has converged
            state.x : the new value of the parameter.
            state.y : the new value of the objective

        """
        if not isinstance(x0, State):
            state = State()

            state.x = x0
            # initialize state with args
            for key, value in state_args.items():
                setattr(state, key, value)

            # check the preconditioner.
            problem.check_preconditioner(x0)
        else:
            state = x0

        optimizer._minimize(problem, state, monitor)

        return state

class VectorSpace(object):
    def __init__(self, addmul=None, dot=None):
        if addmul:
            self.addmul = addmul
        if dot:
            self.dot = dot

    def copy(self, a):
        r = self.addmul(0, a, 1)
        assert type(r) is type(a)
        return r

    def ones_like(self, b):
        r = self.addmul(1, b, 0)
        assert type(r) is type(b)
        return r

    def zeros_like(self, b):
        r = self.addmul(0, b, 0)
        assert type(r) is type(b)
        return r

    def mul(self, b, c, p=1):
        return self.addmul(0, b, c, p)

    def pow(self, c, p):
        i = self.ones_like(c)
        return self.addmul(0, i, c, p)

    def addmul(self, a, b, c, p=1):
        """ Defines the addmul operation.

            either subclass this method or supply a method in the constructor, __init__

            addmul(a, b, c, p) := a + b * c ** p

            The result shall be a vector like b.

            b is always a vector for this VectorSpace; though be aware
            that there can be multiple valid Python types defined on the same
            VectorSpace. For example, particle positions are straight numpy.ndarray,
            An overdensity field may be a ComplexField or a RealField object.
        """

        raise NotImplementedError

    def dot(self, a, b):
        """ defines the inner product operation. 

            dot(a, b) := a @ b

            The result shall be a scalar floating point number.

            a and b are always vector for this VectorSpace, and are guarenteed
            to be of the same Python type -- if not there is a bug from upstream
            callers.

        """
        raise NotImplementedError





from __future__ import print_function
from abopt.legacy.vmad2 import CodeSegment, Engine, statement, programme, ZERO, logger, Literal
from numpy.testing import assert_raises, assert_array_equal, assert_allclose
from numpy.testing.decorators import skipif
import numpy
import logging

try: import graphviz
except ImportError: graphviz = None

class ScalarEngine(Engine):
    def __init__(self):
        pass    

    @statement(ain=['x'], aout=['y'])
    def unitary(engine, x, y, factor):
        y[...] = x * factor

    @unitary.defvjp
    def _(engine, _x, _y, factor):
        _x[...] = _y * factor

    @unitary.defjvp
    def _(engine, x_, y_, factor):
        y_[...] = x_ * factor

    @unitary.jvp.defvjp
    def _(engine, factor, _x_, _y_):
        _x_[...] = _y_ * factor

    @statement(ain=['x'], aout=['y'])
    def power(engine, x, y, factor):
        y[...] = x ** factor

    @power.defvjp
    def _(engine, _x, _y, x, factor):
        if factor > 1:
            _x[...] = _y * ( x ** (factor - 1) * factor if factor > 1 else 1 )
        else:
            x_[...] = _y

    @power.defjvp
    def _(engine, x_, y_, x, factor):
        y_[...] = x_ * x ** (factor - 1) * factor

    @power.jvp.defvjp
    def _(engine, x, factor, _x_, _y_, _x, x_):
        _x[...] = _y_ * x_ * (x ** (factor - 2) * factor * (factor - 1) if factor > 2 else 1)
        _x_[...] = _y_ * ( x ** (factor - 1) * factor if factor > 1 else 1)

def test_scalar():
    engine = ScalarEngine()
    code = CodeSegment(engine)
    code.unitary(x='a', y='r', factor=9.0)
    code.power(x='r', y='b', factor=2.0)
    b = code.compute('b', {'a' : 2.0})
    b, tape = code.compute('b', {'a' : 2.0}, return_tape=True)

    backward_gradient = tape.get_vjp()
    forward_gradient = code.get_jvp()

    b_ = forward_gradient.compute('b_', {'a' : 2.0, 'a_' : 1.0})
    _a = backward_gradient.compute('_a', {'_b' : 1.0})
    assert_array_equal(b_, _a)

    # to do the hessian, first augment the compute with a forward pass 

    # then do the backward gradient on the forward tape
    # we need us terms like x_, etc.

    print(forward_gradient)
    (b_, b), tape = forward_gradient.compute(['b_', 'b'], {'a' : 2.0, 'a_' : 1.0}, return_tape=True)
    print('b_', 'b', b_, b)
    hessian_dot = tape.get_vjp()
    print(hessian_dot)

    _a = hessian_dot.compute('_a', {'_b_' : 1.0, '_b' : 0.0}, monitor=lambda node, frontier, r:print('---', node, frontier, r))
    print(_a)

class VectorEngine(Engine):
    def __init__(self):
        pass    

    @statement(ain=['x'], aout=['y'])
    def multiply(engine, x, y, factor):
        y[...] = x * factor

    @multiply.defvjp
    def _(engine, _x, _y, factor):
        _x[...] = _y * factor

    @multiply.defjvp
    def _(engine, x_, y_, factor):
        y_[...] = x_ * factor

    @multiply.jvp.defvjp
    def _(engine, factor, _y_, _x_):
        _x_[...] = _y_ * factor

    @statement(ain=['x1', 'x2'], aout=['y'])
    def dot(engine, x1, x2, y):
        y[...] = numpy.einsum('i,i->', x1, x2)

    @dot.defvjp
    def _(engine, _x1, _x2, x1, x2, _y):
        _x1[...] = _y * x2
        _x2[...] = _y * x1

    @dot.defjvp
    def _(engine, x1_, x2_, x1, x2, y_):
        y_[...] = numpy.einsum('i,i->', x1, x2_) \
                + numpy.einsum('i,i->', x2, x1_)

    @dot.jvp.defvjp
    def _(engine, x1_, x2_, x1, x2, _x1_, _x2_, _x1, _x2, _y_):
        _x1_[...] = x2 * _y_
        _x2_[...] = x1 * _y_
        _x1[...] = _y_ * x2
        _x2[...] = _y_ * x1

    @statement(ain=['x'], aout=['y'])
    def power(engine, x, y, factor):
        y[...] = x ** factor

    @power.defvjp
    def _(engine, _x, _y, x, factor):
        _x[...] = _y * (x ** (factor - 1) * factor if factor != 1 else 1)

    @power.defjvp
    def _(engine, x_, y_, x, factor):
        y_[...] = x_ * (x ** (factor - 1) * factor if factor != 1 else 1)

    @power.jvp.defvjp
    def _(engine, x_, x, factor, _x_, _y_, _x):
        _x_[...] = _y_ * (x ** (factor - 1) * factor if factor != 1 else 1)
        _x[...] = _y_ * x_ * (x ** (factor - 2) * (factor  - 1) * factor if factor != 2 else 1)

def impulse(shape, ind):
    x = numpy.zeros(shape)
    x[ind] = 1.0
    return x

def bases(shape, return_index=False):
    for ind in numpy.ndindex(*shape):
        if return_index:
            yield impulse(shape, ind), ind
        else:
            yield impulse(shape, ind)

def test_vector():
    engine = VectorEngine()
    code = CodeSegment(engine)

    code.multiply(x='x1', y='r1', factor=3.0)
    code.multiply(x='x2', y='r2', factor=3.0)
    code.dot(x1='r1', x2='r2', y='r3')
    code.power(x='r3', y='y', factor=2)

    init = {'x1' : numpy.array([1.0, 1.0]),
            'x2' : numpy.array([1.0, 1.0])}
    b, tape = code.compute('y', init, return_tape=True)

    backward_gradient = tape.get_vjp()
    forward_gradient = code.get_jvp()

    _x1, _x2 = backward_gradient.compute(['_x1', '_x2'], {'_y' : 1.0})

    for (x1_, x2_), ind in bases((2,2), return_index=True):
        d = {'x1_' : x1_, 'x2_' : x2_}
        d.update(init)
        y_ = forward_gradient.compute('y_', d)
        assert_array_equal(y_ , numpy.array([_x1, _x2])[ind])

    x1_, x2_ = impulse((2, 2), (0, 1))

    d = {'x1_' : x1_, 'x2_' : x2_}
    d.update(init)
    (y, y_), tape = forward_gradient.compute(['y', 'y_'], d, return_tape=True)
    hessian_dot = tape.get_vjp()
    _x1, _x2 = hessian_dot.compute({'_x1', '_x2'}, {'_y_' : 1.0, '_y' : 0.0})




from __future__ import print_function
import warnings
import functools
import logging

logger = logging.getLogger("VMAD")
_logging_handler = logging.StreamHandler()
logger.addHandler(_logging_handler)

# TODO:
# Add visualization

def ZeroType():
    """ creates a special type of ZeroType; """
    def self(self, *args): return self
    def other(self, other): return other
    def zde(self, a): raise ZeroDivisionError
    def __sub__(self, a): return -a
    def __xor__(self, a): return ~a
    def __int__(self): return 0
    def __float__(self): return 0.0
    def __round__(self): return 0
    def __array__(self, dtype=None):
        import numpy
        return numpy.array(0, dtype=dtype)
    def __repr__(self): return "<ZERO>"
    dict = {}
    for name, value in locals().items():
         if name.startswith("__"): dict[name] = value

    for name in [
        "neg", "pos", "abs", "invert", "complex",
        "mul", "rmul", "matmul", "rmatmul", 
        "mod", "divmod", "div", "truediv", "floordiv",
        "pow",
        "and", "rand", "lshift", "rlshift", "rshift", "rrshift",
        "getitem", "reversed"]:
        dict["__" + name + "__"] = self

    for name in [
        "rmod", "rdivmod", "rdiv", "rtruediv", "rfloordiv",
        "rpow", "rsub", "rxor"]:
        dict["__" + name + "__"] = zde

    for name in [
        "add", "radd", "or", "ror"]:
        dict["__" + name + "__"] = other

    dict["__repr__"] = __repr__
    return type("ZeroType", (object,), dict)

ZeroType = ZeroType()
ZERO = ZeroType()

# decorators
def statement(ain, aout): return lambda body: Statement(body, ain, aout)
def programme(ain, aout): return lambda body: Programme(body, ain, aout)

# create a model using a given computing engine
def model(engine): return CodeSegment(engine)

class LValue(object):
    def __init__(self, name, ns):
        self.ns = ns
        self.name = name

    def __getattr__(self, attr): return getattr(self[...], attr)
    def __repr__(self): return "LValue:%s" % self.name
    def __getitem__(self, index): return self.ns[self.name]
    def __setitem__(self, index, value): self.ns[self.name] = value

class Literal(object):
    def __init__(self, value):
        self.value = value
    def __repr__(self): return "Literal:%s" % _short_repr(self.value)

class Primitive(object):
    def __init__(self, body, ain, aout, argnames=None):
        self.body = body
        self.ain = ain
        self.aout = aout
        if argnames is None:
            argnames = body.__code__.co_varnames[1:body.__code__.co_argcount]

        if getattr(body, '__kwdefaults__', None) is not None:
            self.defaults = dict(body.__kwdefaults__)
        elif getattr(body, '__defaults__', None) is not None:
            self.defaults = dict(zip(argnames[-len(body.__defaults__):], body.__defaults__))
        else:
            self.defaults = {}

        self.argnames = argnames
        for an in ain:
            if not an in self.argnames:
                raise ValueError(
    "argument `%s` of ain in microcode decorator is not declared by function `%s`"
                       % (an, str(self.body))
                )
        functools.update_wrapper(self, body)

        args = Arguments()
        for argname in argnames:
            if argname in ain:
                if argname in aout:
                    arg = IOArgument(argname)
                else:
                    arg = IArgument(argname)
            elif argname in aout:
                arg = OArgument(argname)
            else:
                arg = EXArgument(argname)
            args.append(arg)

        self.args = args
        body.args = args

    def __repr__(self):
        return self.body.__name__

    def create_node(self, engine):
        nodetype = type(self).NodeType
        return nodetype(engine, self)

class Variable(object):
    """ if the same variable name is modified we use a different postifx
        this happens as Variable mentioned in Code as O/IOArgument.
    """
    def __init__(self, name, postfix):
        self.name = name
        self.postfix = postfix

    @property
    def name_vjp(self): return '_' + self.name

    @property
    def name_jvp(self): return self.name + '_'

    def __hash__(self): return hash(self.name + '-%s' % self.postfix)
    def __eq__(self, other): return self.name == other.name and self.postfix == other.postfix

    def __repr__(self):
        if self.postfix is not None:
            return "%s/%d" % (self.name, self.postfix)
        else:
            return "%s" % (self.name)

class Argument(object):
    def __init__(self, name):
        self.name = name
        self.value = None
        self.ovalue = None

    def copy(self):
        arg = type(self)(self.name)
        arg.value = self.value
        arg.ovalue = self.ovalue
        return arg

    @property
    def name_vjp(self): return '_' + self.name

    @property
    def name_jvp(self): return self.name + '_'


    def dereference(self, context):
        """ returns the value of an argument by its value

            if context is None, returns the name of the variable
        """
        if isinstance(self.value, Literal):
            if context is None:
                return self.value
            else:
                return self.value.value
        elif isinstance(self.value, Variable):
            if context is None:
                return self.value.name
            else:
                return context[self.value.name]
        else:
            return self.value

    def __repr__(self):
        if isinstance(self, IOArgument):
            return "%s:%s=%s=>%s" % (type(self).__name__, self.name, self.value, self.ovalue)
        else:
            return "%s:%s=%s" % (type(self).__name__, self.name, _short_repr(self.value))

class IArgument(Argument): pass
class OArgument(Argument): pass
class IOArgument(Argument): pass
class EXArgument(Argument): pass
class Arguments(list):
    def copy(self):
        args = Arguments()
        for arg in self:
            args.append(arg.copy())
        return args

    def find(self, argname):
        for arg in self:
            if arg.name == argname: return arg
        else:
            raise KeyError

    def get_kwargs(self):
        kwargs = {}
        for arg in self:
            if isinstance(arg.value, Variable):
                kwargs[arg.name] = arg.value.name
            else:
                kwargs[arg.name] = arg.value
        return kwargs

    def set_values(self, kwargs, defaults, code):
        _kwargs = defaults.copy()
        _kwargs.update(kwargs)
        kwargs = _kwargs

        for arg in self:
            if isinstance(arg, EXArgument):
                variable = kwargs.pop(arg.name)
                arg.value = variable
        for arg in self:
            if isinstance(arg, IArgument):
                varname = kwargs.pop(arg.name, arg.name)
                variable = code.get_latest_variable(varname, expired=False)
                arg.value = variable
        for arg in self:
            if isinstance(arg, IOArgument):
                varname = kwargs.pop(arg.name, arg.name)
                variable = code.get_latest_variable(varname, expired=True)
                arg.value = variable
                arg.ovalue = code.create_latest_variable(varname)
        for arg in self:
            if isinstance(arg, OArgument):
                varname = kwargs.pop(arg.name, arg.name)
                variable = code.create_latest_variable(varname)
                arg.ovalue = variable
                arg.value = variable

        if len(kwargs) > 0:
            raise ValueError("additional kwargs are found: %s" % list(kwargs.keys()))

class Node(object):
    # if true, invoke will directly return ZERO when all inputs are ZERO
    ZERO_BYPASS = False 

    def __init__(self, engine, primitive):
        self.primitive = primitive
        self.engine = engine
        self.args = primitive.args.copy()

    def copy(self):
        return type(self)(self.engine, self.primitive)

    def bind(self, frontier, results):
        """ bind args to objects in frontier, or LValues """
        bound = []
        primitive = self.primitive
        for arg in self.args:
            if isinstance(arg, (IArgument, IOArgument)):
                bound.append(arg.dereference(frontier))
            elif isinstance(arg, OArgument):
                bound.append(LValue(arg.value.name, results))
            else:
                bound.append(arg.value)
        return bound

    def __repr__(self):
        return "%s(%s)" % (self.primitive, self.args)

    def call(self, bound, return_tape=False):
        # a simple node doesn't know how to return a Tape
        assert not return_tape
        r = self.primitive.body(self.engine, *bound)
        return None

    def invoke(self, frontier):
        #logger.info("Invoke %s" % (self))
        out = {}
        bound = self.bind(frontier, out)

        if self.ZERO_BYPASS and (
            all([value is ZERO for arg, value in zip(self.args, bound) 
                if isinstance(arg, (IArgument, IOArgument))])):

            for arg, value in zip(self.args, bound):
                if isinstance(arg, OArgument):
                    # IOArguments are already ZEROs
                    value[...] = ZERO
        else:
            self.call(bound)

        for arg, value in zip(self.args, bound):
            if isinstance(arg, IOArgument):
                out[arg.value.name] = value

        return out

    def invoke_for_tape(self, frontier):
        bound = self.bind(frontier, {})
        tape = self.call(bound, return_tape=True)
        # The call didn't return a Tape. Likley calling the method on a statement's Node?
        assert isinstance(tape, Tape)
        return tape

class CodeSegNode(Node):
    def get_codeseg(self):
        raise NotImplementedError

    def call(self, bound, return_tape=False):
        init = {}
        lvalues = {}
        for arg, value in zip(self.args, bound):
            if isinstance(arg, (IArgument, IOArgument)):
                init[arg.name] = value
            elif isinstance(arg, (EXArgument,)):
                init[arg.name] = value
            else:
                lvalues[arg.name] = value

        aout = [ arg.name for arg in self.args
                if isinstance(arg, (OArgument, IOArgument))]

        codeseg = self.get_codeseg()

        # compute doesn't taint init.
        out = codeseg.compute(aout, init, return_tape=return_tape)
        #logger.info("CodeSegment results %s %s" % (aout, _short_repr(out)))
        if return_tape:
            out, tape = out
        else:
            tape = None

        for argname, value in zip(aout, out):
            lvalues[argname][...] = value

        return tape

class Statement(Primitive):
    NodeType = Node

    def __init__(self, body, ain, aout):
        Primitive.__init__(self, body, ain, aout)

    def defvjp(self, body):
        """ Define the back-propagation gradient operator. """
        gout = ['_' + a for a in self.ain]
        gin  = ['_' + a for a in self.aout]

        argnames = body.__code__.co_varnames[1:body.__code__.co_argcount]
        ain = [a for a in self.ain if a in argnames]

        body.__name__ = "BG:" + self.body.__name__
        self.vjp = StatementVJP(body, gin + ain, gout)

        # allow the gradient with the same name as the original body.
        return self.vjp

    def defjvp(self, body):
        """ Define the forward-propagation gradient operator. """
        gin = [a + '_' for a in self.ain]
        gout  = [a + '_' for a in self.aout]
        argnames = body.__code__.co_varnames[1:body.__code__.co_argcount]
        ain = [a for a in self.ain if a in argnames]
        body.__name__ = "FG:" + self.body.__name__

        self.jvp = StatementJVP(body, gin + ain, gout)

        # allow the gradient with the same name as the original body.
        return self.jvp

class StatementVJP(Statement):
    class NodeType(Node):
        ZERO_BYPASS = True

class StatementJVP(Statement):
    class NodeType(Node):
        ZERO_BYPASS = True

class Programme(Primitive):
    def __init__(self, body, ain, aout):
        Primitive.__init__(self, body, ain, aout)
        self.vjp = ProgrammeVJP(self)
        self.jvp = ProgrammeJVP(self)

    class NodeType(CodeSegNode):
        def get_codeseg(self):
            return self.primitive.body(self.engine,
                *[arg.value if isinstance(arg, EXArgument)
                else arg.name for arg in self.args])

class ProgrammeVJP(Primitive):
    def __init__(self, programme):
        gout = ['_' + a for a in programme.ain]
        gin  = ['_' + a for a in programme.aout]
        ex = [a for a in programme.argnames if a not in programme.aout + programme.ain]
        extra = ['#replay-record']
        body = lambda : None
        body.__name__ = "VJP:" + programme.body.__name__
        argnames = list(set(gin + gout + programme.ain + ex + extra))
        Primitive.__init__(self, body, gin, gout, argnames=argnames)

    class NodeType(CodeSegNode):
        ZERO_BYPASS = True

        def get_codeseg(self):
            # replay then obtain the gradient codeseg
            node, d = self.args.find('#replay-record').value

            tape = node.invoke_for_tape(d)

            vjpcode = tape.get_vjp()
            # if an output variable is not mentioned in the code
            # then the vjp code segment doesn't set the default to ZERO
            # we fix it here.
            for arg in self.args:
                if isinstance(arg, OArgument):
                    vjpcode.defaults[arg.name] = ZERO

            return vjpcode

class ProgrammeJVP(Primitive):
    def __init__(self, programme):
        gin  = [a + '_' for a in programme.ain]
        gout = [a + '_' for a in programme.aout]
        ex = [a for a in programme.argnames if a not in programme.aout + programme.ain]
        extra = ['#replay-record']
        body = lambda : None
        body.__name__ = "JVP:" + programme.body.__name__
        argnames = list(set(gin + gout + programme.ain + ex + extra))
        Primitive.__init__(self, body, gin, gout, argnames=argnames)

    class NodeType(CodeSegNode):
        ZERO_BYPASS = True

        def get_codeseg(self):
            node, d = self.args.find('#replay-record').value
            tape = node.invoke_for_tape(d)
            # Watch out: We use the tape version
            # of VJP because with the code version of VJP
            # we do not know how to pass in the arguments
            # these arguments are marked as EXArgument in the VJP
            # but we need to resolve them from the frontier.

            jvpcode = tape.get_jvp()
            return jvpcode

class Tape(object):
    def __init__(self, engine, init):
        self.records = []
        self.init = {}
        self.init.update(init)
        self.engine = engine

    def append(self, node, frontier):
        d = {}
        for arg in node.args:
            # remember all input variable as their values
            if isinstance(arg, (IArgument, IOArgument)) \
                 and isinstance(arg.value, Variable):
                d[arg.value.name] = arg.dereference(frontier)

        self.records.append((node, d))

    def __repr__(self):
        return '\n'.join('%s | %s' % (node, list(d.keys())) for node, d in self.records)

    def get_vjp(self):
        """ Create a code segment that computes the vector jacobian product for a tape,
            with backward gradient propagation.

            A vector jacobian product is J_ij v_j where j is the output variable index.

            The input variable of the returned CodeSegment is '_a', '_b', ... where a, b,
            ... are the output variables.
        """

        code = CodeSegment(self.engine)

        add = self.engine.add

        first_time = {} 
        for node, d in self.records[::-1]:
            vjp = node.primitive.vjp
            kwargs = {}
            partials = []
            for arg in node.args:
                if isinstance(arg, OArgument) \
                and arg.name_vjp in vjp.argnames:
                    kwargs[arg.name_vjp] = arg.value.name_vjp
                if isinstance(arg, (IArgument, IOArgument)) \
                and arg.name in vjp.argnames:
                    kwargs[arg.name] = Literal(arg.dereference(d))

                if isinstance(arg, (IArgument, IOArgument)) and \
                   arg.name_vjp in vjp.argnames:
                    if isinstance(arg.value, Literal):
                        kwargs[arg.name_vjp] = '###abandon###'
                    else:
                        if first_time.get(arg.value, True):
                            # directly write to the gradient, it is used
                            kwargs[arg.name_vjp] = arg.value.name_vjp
                            first_time[arg.value] = False
                        else:
                            newname = arg.value.name_vjp + '#partial'
                            kwargs[arg.name_vjp] = newname
                            partials.append((newname, arg.value.name_vjp))

                if isinstance(arg, EXArgument) and arg.name in vjp.argnames:
                    kwargs[arg.name] = arg.value

            if isinstance(node.primitive, Programme):
                # the vjp of a Programme requires more arguments
                # to build the vjp codesegment on the fly
                kwargs['#replay-record'] = (node, d)

            code.append(vjp, kwargs)
            for p, r in partials:
                kwargs = {}
                kwargs['x1'] = p
                kwargs['x2'] = r
                kwargs['y'] = r
                code.append(add, kwargs)

        for variable in code._input_variables.values():
            code.defaults[variable.name] = ZERO

        #logger.info("GRADIENT code.defaults: %s " % code.defaults)
        return code

    def get_jvp(self):
        """ creates a CodeSegment that computes the jacobian vector product, from a tape,
            via forward gradient propagation.

            A jacobian vector product is J_ij v_i where i is index of the input variables.

            The returned CodeSegment input is 'a_', 'b_', ... where 'a', 'b', ...
            are the input variables of the original code segment.

            The advantage of starting from a tape is that we do not need to compute
            the original code together with the forward pass. Useful if we need to do
            vjp and jvp same time.
        """
        code = CodeSegment(self.engine)

        for node, d in self.records:
            jvp = node.primitive.jvp
            kwargs = {}
            for arg in node.args:
                if isinstance(arg.value, Variable) and arg.name_jvp in jvp.argnames:
                    kwargs[arg.name_jvp] = arg.value.name_jvp
                if isinstance(arg, IArgument) and arg.name in jvp.argnames:
                    kwargs[arg.name] = Literal(arg.dereference(d))
                    # for literal inputs, we shall set the input gradient to zero
                    if isinstance(arg.value, Literal):
                        kwargs[arg.name_jvp] = ZERO
                if isinstance(arg, IOArgument) and arg.name in jvp.argnames:
                    kwargs[arg.name] = Literal(arg.dereference(d))
                if isinstance(arg, EXArgument) and arg.name in jvp.argnames:
                    kwargs[arg.name] = arg.value

            if isinstance(jvp, ProgrammeJVP):
                kwargs['#replay-record'] = node, d

            code.append(jvp, kwargs)

        for variable in code._input_variables.values():
            code.defaults[variable.name] = ZERO

        return code

    def to_graph(self, **kwargs):
        nodes = [node for node, kwargs in self.records]
        return nodes_to_graph(nodes, **kwargs)[0]

class CodeSegment(object):
    def __init__(self, engine):
        self.engine = engine
        self.nodes = []
        self.defaults = {} # use these if not provided in init

        self._liveset = {} # the set of variables ready to be used as input (latest variables).
                           # If a variable is destroyed the value replaced with None.

        self._postfix = 0 # a unique postfix added to every variable;

        # if a variable is used as input but not yet been mentioned on the liveset.
        self._input_variables = {} # input variables

    def __getattr__(self, name):
        """ Allow looking up primitives and programmes from the engine namespace """
        try:
            item = getattr(self.engine, name)
        except AttributeError:
            raise AttributeError("%s is not a declared primitiveuction in %s" % (name, type(self.engine)))

        if isinstance(item, Primitive):
            primitive = item
            def func(**kwargs):
                self.append(primitive, kwargs)
            functools.update_wrapper(func, item)
            return func
        else:
            raise TypeError

    def __dir__(self):
        l = []
        for name in dir(self.engine):
            item = getattr(self.engine, name)
            if isinstance(item, Primitive):
                l.append(name)
        return l + dir(type(self))

    def copy(self):
        code = CodeSegment(self.engine)
        code.nodes.extend(self.nodes)
        code.defaults.update(self.defaults)
        code._liveset.update(self._liveset)
        code._postfix = self._postfix
        code._input_variables = self._input_variables
        return code

    def get_latest_variable(self, varname, expired=False):
        if isinstance(varname, Literal):
            return varname
        if varname not in self._liveset:
            self._postfix = self._postfix + 1
            variable = Variable(varname, self._postfix)
            self._input_variables[varname] = variable
        else:
            variable = self._liveset.get(varname)
            if variable is None:
                self._postfix = self._postfix + 1
                variable = Variable(varname, self._postfix)

        if not expired:
            self._liveset[varname] = variable
        else:
            self._liveset[varname] = None

        return variable

    def create_latest_variable(self, varname):
        self._postfix = self._postfix + 1
        variable = Variable(varname, self._postfix)
        self._liveset[varname] = variable
        return variable

    def append(self, primitive, kwargs):
        node = primitive.create_node(self.engine)
        node.args.set_values(kwargs, primitive.defaults, self)
        self.nodes.append(node)

    def optimize(self, vout):
        out = [ self.get_latest_variable(varname) for varname in vout]
        nodes = _optimize(self.nodes, out)
        segment = self.copy()
        segment.nodes = nodes
        return segment

    def compute(self, vout, init, return_tape=False, monitor=None):
        assign = self.engine.assign.body

        if not isinstance(vout, (tuple, list, set)):
            vout = [vout]
            squeeze = True
        else:
            squeeze = False

        frontier = {}

        for var, value in self.defaults.items():
            frontier[var] = value
        for var, value in init.items():
            frontier[var] = value

        if return_tape:
            tape = Tape(self.engine, frontier)

            # XXX This is not nice. But requires too much 
            # refactoring to get it right.
            self = self.copy()
            # we need to connect the outputs to a 'terminal'
            # node such that their input gradients are not
            # overwritten by the partial gradients of
            # subsequential operations.
            @statement(ain=['x'], aout=['x'])
            def mark(engine, x): pass
            @mark.defvjp
            def _(engine, _x): pass
            @mark.defjvp
            def _(engine, x_): pass
            @mark.vjp.defjvp
            def _(engine, _x_): pass
            for varname in vout:
                self.append(mark, {'x' : varname})
        else:
            tape = None

        out = [ self.get_latest_variable(varname) for varname in vout]

        nodes = _optimize(self.nodes, out)
        freeables = _get_freeables(nodes, out)

        for i, (node, abandon) in enumerate(zip(nodes, freeables)):
            if tape:
                tape.append(node, frontier)
                if node.primitive is not mark:
                    for arg in node.args:
                        if not isinstance(arg, IOArgument): continue
                        # FIXME: use copy
                        assign(self.engine, x=frontier[arg.value.name], y=LValue(arg.value.name, frontier))
            try:
                r = node.invoke(frontier)
            except Exception as e:
                print("Failure in running `%s`" % node)
                raise

            if monitor is not None:
                monitor(node, frontier, r)
            for var in abandon:
                frontier.pop(var.name)

#            if len(abandon):
#                logger.info("Removed from frontier %s", abandon)
            frontier.update(r)
            #logger.info("Frontier %s", list(frontier.keys()))

        r = [frontier[vn] for vn in vout]
        if squeeze:
            r = r[0]
        if return_tape:
            r = r, tape
        return r

    def get_jvp(self, init={}):
        """ creates a CodeSegment that computes the jacobian vector product, with forward
            gradient propagation.

            A jacobian vector product is J_ij v_i where i is index of the input variables.

            The returned CodeSegment input is 'a_', 'b_', ... where 'a', 'b', ...
            are the input variables of the original code segment.

            This will compute the original code together with the forward gradient pass.

        """
        code = CodeSegment(self.engine)

        for node in self.nodes:
            jvp = node.primitive.jvp
            kwargs = {}
            for arg in node.args:
                if isinstance(arg.value, Variable) and arg.name_jvp in jvp.argnames:
                    kwargs[arg.name_jvp] = arg.value.name_jvp
                if isinstance(arg, (IArgument, IOArgument)) and arg.name in jvp.argnames:
                    kwargs[arg.name] = arg.dereference(None)
                if isinstance(arg, EXArgument) and arg.name in jvp.argnames:
                    kwargs[arg.name] = arg.value

            if isinstance(jvp, ProgrammeJVP):
                # empty init because all variables are on the frontier.
                kwargs['#replay-record'] = node, {}

            code.append(jvp, kwargs)
            code.append(node.primitive, node.args.get_kwargs())

        for variable in code._input_variables.values():
            code.defaults[variable.name] = ZERO

        # merge in the defaults of self
        code.defaults.update(self.defaults)

        # initialize with the defaults
        code.defaults.update(init)
        return code

    def compute_with_gradient(self, vout, init, ginit, return_tape=False):
        if not isinstance(vout, (tuple, list, set)):
            vout = [vout]
            squeeze = True
        else:
            squeeze = False

        cnout = [vn for vn in vout if not vn.startswith('_')]
        # if gradient request are requested, they must be computed
        cnout_g = [ vn[1:] for vn in ginit]

        gnout = [vn for vn in vout if vn.startswith('_')]

        cout, tape = self.compute(cnout + cnout_g, init, return_tape=True)
        cout = cout[:len(cnout)]

        vjpcode = tape.get_vjp()

        gout = vjpcode.compute(gnout, ginit)

        d = {}
        d.update(zip(cnout, cout))
        d.update(zip(gnout, gout))

        out = [d[vn] for vn in vout]
        if squeeze:
            out = out[0]
        return out

    def __repr__(self):
        nodes = '\n'.join('%s' % node for node in self.nodes)
        return '\n'.join([nodes])

    def to_graph(self, **kwargs):
        return nodes_to_graph(self.nodes, **kwargs)[0]

# an optional base class for computing engines
class Engine(object):
    @statement(ain=['x1', 'x2'], aout=['y'])
    def add(engine, x1, x2, y):
        y[...] = x1 + x2

    @statement(ain=['x'], aout=['y'])
    def assign(engine, x, y):
        y[...] = x * 1.0

def _optimize(nodes, out):
    """ return an optimized codeseg for computing vout. irrelevant nodes are pruned. """

    deps = set(out)

    newnodes = []
    for node in nodes[::-1]:
        keep = False
        for arg in node.args:
            if isinstance(arg, OArgument) and arg.value in deps:
                keep = True
                deps.remove(arg.value)
            if isinstance(arg, IOArgument) and arg.ovalue in deps:
                keep = True
                deps.remove(arg.ovalue)
        if not keep: continue
        newnodes.append(node)
        for arg in node.args:
            if isinstance(arg, (IArgument, IOArgument)):
                deps.add(arg.value)

    return list(reversed(newnodes))

def _get_freeables(nodes, out):
    refcounts = {}
    for var in out:
        refcounts[var] = refcounts.get(var, 0) + 1

    for node in nodes:
        for arg in node.args:
            if isinstance(arg, (IArgument, IOArgument)):
                refcounts[arg.value] = refcounts.get(arg.value, 0) + 1

    free_list = []

    for node in nodes:
        item = []
        for arg in node.args:
            if isinstance(arg, (IArgument, IOArgument)):
                if isinstance(arg.value, Literal): continue
                refcounts[arg.value] = refcounts[arg.value] - 1
                if refcounts[arg.value] == 0:
                    item.append(arg.value)
        free_list.append(item)
    return free_list

def nodes_to_graph(nodes, depth=0, **kwargs):
    """
        Graph representation of nodes, kwargs are sent to graphviz

        depth controls the behavior of programme nodes. only 
        depth level of sub graphs are made.
    """
    import graphviz
    graph = graphviz.Digraph(**kwargs)

    def unique(obj):
        return '%08X%08X' % (id(nodes), id(obj))

    subgraphs = {}

    for i, node in enumerate(nodes):
        label = '%s<BR/>' % str(node.primitive)
        ex = []
        for arg in node.args:
            if isinstance(arg, EXArgument):
                # bypass aux arguments starting with sharp
                if not arg.name.startswith('#'):
                    ex.append(str(arg))
        label = label + '<BR/>'.join(ex)
        label = '<' + label + '>'

        if depth > 0 and isinstance(node, CodeSegNode):
            # render the programme nodes as subgraphs
            
            codeseg = node.get_codeseg()
            subgraph, inputs, outputs = nodes_to_graph(codeseg.nodes, depth - 1)
            subgraph.name = 'cluster_' + str(node.primitive)
            subgraph.attr('graph', label=label)
            subgraph.attr('graph', color='blue')
            subgraph.attr('graph', style='dotted')
            graph.subgraph(subgraph)
            subgraphs[unique(node)] = (inputs, outputs)
        else:
            graph.node(unique(node), label=label, shape='box')

    source = {}

    inputs, outputs = {}, {}

    def process_in_arg(arg, node):
        attrs = {}
        attrs['label'] = '<' + str(arg.value) + '>'
        nodeid = unique(node)
        if isinstance(arg, (IArgument, IOArgument)):
            if arg.value in source:
                from_nodeid, from_arg = source[arg.value]
                attrs['taillabel'] = '<' + str(from_arg.name) + '>'
                attrs['tail_lp'] = "12"
            else:
                from_nodeid = nodeid + unique(arg.value)
                graph.node(from_nodeid, label=str(arg.value))

                if not isinstance(arg.value, Literal):
                    inputs[arg.value.name] = from_nodeid
                          
            if nodeid in subgraphs:
                nodeid = subgraphs[nodeid][0][arg.name]

            attrs['headlabel'] = '<' + str(arg.name) + '>'
            attrs['head_lp'] = "12"
            graph.edge(from_nodeid, nodeid, **attrs)

    def process_out_arg(arg, node):
        if isinstance(arg, (OArgument, IOArgument)):
            nodeid = unique(node)
            if nodeid in subgraphs:
                nodeid = subgraphs[nodeid][1][arg.name]

            source[arg.ovalue] = (nodeid, arg)
            outputs[arg.value.name] = nodeid

    for i, node in enumerate(nodes):
        for arg in node.args:
            process_in_arg(arg, node)
            process_out_arg(arg, node)

    return graph, inputs, outputs


def _short_repr(obj):
    if isinstance(obj, (list, tuple)):
        return [_short_repr(i) for i in obj]
    else:
        s = '%s' % obj
        if len(s) > 30:
            s = '[%s]' % type(obj).__name__
        return s



from __future__ import print_function
import warnings
import functools
class LValue(object):
    def __init__(self, vn, ns):
        self.ns = ns
        self.vn = vn

    def __getattr__(self, attr):
        return getattr(self[...], attr)

    def __repr__(self):
        return "LValue:%s" % self.vn

    def __getitem__(self, index):
        return self.ns[self.vn]

    def __setitem__(self, index, value):
#        assert index in (0, Ellipsis, None)
        # anything should work -- to support [:]
        self.ns[self.vn] = value

class MicroCode(object):
    """ An object that represents a microcode.

        Mostly a python function with some additional introspection, marking
        the input and output variables.
    """
    def __init__(self, function, ain, aout, is_programme=False):
        self.function = function
        self.ain = ain
        self.aout = aout
        self.argnames = function.__code__.co_varnames[1:function.__code__.co_argcount]
        for an in ain:
            if not an in self.argnames:
                raise ValueError(
    "argument `%s` of ain in microcode decorator is not declared by function `%s`"
                       % (an, str(self.function))
                )
        self.vjp = NotImplemented
        self.is_programme = is_programme
        functools.update_wrapper(self, function)

    def defvjp(self, function):
        """ Define the back-propagation gradient operator. """
        gout = ['_' + a for a in self.ain]
        gin  = ['_' + a for a in self.aout]

        function.__name__ = "G:" + self.function.__name__
        self.vjp = microcode(gin, gout)(function)
        # allow the gradient with the same name as the original function.
        return self.vjp
    grad = defvjp

    def __get__(self, instance, owner):
        """ As a class member, return the microcode,
            As an instance member of VM, returns the function as a method,
            As an instance member of Code, returns a method to add to the code.
        """
        if instance is not None:
            if self.is_programme:
                # programme is directly ran.
                @functools.wraps(self.function)
                def method(_self, *args, **kwargs):
                    for a in self.ain + self.aout:
                        kwargs.setdefault(a, a)
                    return self.function(_self, *args, **kwargs)
                return method.__get__(instance, owner)
            else:
                if isinstance(instance, Code):
                    @functools.wraps(self.function)
                    def method(**kwargs):
                        instance.append(self, kwargs)
                    return method
                else:
                    return self.function.__get__(instance, owner)
        else:
            # class member
            return self

    def __repr__(self):
        return self.function.__name__

    def to_label(self, kwargs, d={}, html=False):
        if html:
            delim = '<BR ALIGN="CENTER"/>'
        else:
            delim = " "
        def format(kwargs, d):
            r = []
            for an in self.argnames:
                vn = kwargs.get(an, an)
                uid, value = d.get(an, (-1, None))
                if an not in self.ain + self.aout:
                    vn = "Ext:%s" % type(vn).__name__
                    io = 'p'
                else:
                    io = ''
                    if an in self.ain:
                        io += 'i'
                    if an in self.aout:
                        io += 'o'
                if uid == -1:
                    s = "%s:%s = %s" % (io, an, vn)
                else:
                    s = "%s:%s = %s [%08X]" % (io, an, vn, uid)
                r.append(s)
            return delim.join(r)
        name = str(self)
        if html:
            name = "<b>" + name + "</b>"

        s = delim.join([name, format(kwargs, d)])
        if html:
            return "<" + s + ">"
        else:
            return s

    def invoke(self, vm, frontier, kwargs, tape, monitor=None):
        vin = []
        tapein = []
        dout = {}
        oldkwargs = kwargs
        kwargs = kwargs.copy()

        # copy in all arguments
        for an in self.argnames:
            if an in self.ain:
                # replace argument with variable name
                # then fetch it from the frontier.
                vn = kwargs.pop(an, an)
                if vn not in frontier:
                    raise ValueError("Argument `%s' of `%s` resolves to an undefined varabile `%s'" % (an, self, vn))
                data = frontier[vn]

                tapein.append(data)

                if an in self.aout:
                    # create a copy, store to the output dict, for tainting the values.
                    # FIXME: only needed if we are recording to a tape.
                    lv = LValue('data', {})
                    vm.CopyVariable(data, lv)
                    data = lv.ns['data']
                    # the value itself must allow inplace operations.
                    dout[vn] = data
                vin.append(data)
            elif an in self.aout:
                vn = kwargs.pop(an, an)
                # out args, make a LValue for [...] assignments
                # mixed inout is handled in above;
                data = LValue(vn, dout)
                vin.append(data)
                tapein.append(data)
            else:
                # use the kwargs -- raise error if not found!
                if an not in kwargs:
                    raise ValueError("Argument `%s' of `%s' could not be bound to a keyword argument" % (an, self))
                data = kwargs.pop(an)
                vin.append(data)
                tapein.append(data)

        if len(kwargs) > 0:
            raise ValueError("Bad keyword arguments : %s" % (','.join(kwargs.keys())))

        if tape is not None:
            tape.record(self, oldkwargs, tapein)

        self.function(vm, *vin)

        if monitor:
            din = dict(zip(self.argnames, vin))
            monitor(self, din, dout, frontier)
        return dout

def programme(ain, aout):
    def decorator(func):
        return MicroCode(func, ain, aout, is_programme=True)
    return decorator
    
def microcode(ain, aout):
    """ Declares a VM member function as a 'microcode'.
        microcode is the building block for Code objects,
        which can be computed and differentiated.

        See MicroCode. 
    """
    def decorator(func):
        return MicroCode(func, ain, aout)
    return decorator

class VM(object):
    """ A virtual machine that interprets and runs Code objects
        consisting of microcodes.

        Subclass VM to add more microcodes.
        Override `CopyVariable` and `Add` to support different types of
        variables.

        Convention for gradient is to prepend `_`, e.g. `_x` is the gradient
        backtraced to `x`.

        The Each microcode carries a list of input and output arguments (
        `ain` and `aout`) that can be assigned as variable names (string).

        The rest of arguments in the function signature
        are external parameters who do not go into gradients.

        Example
        -------

        >>> vm = VM()
        >>> code = vm.code()
        >>> code.Add(a='x', b='x', c='y')
        >>> code.compute('y', {'x' : 10})

    """

    @microcode(ain=['x'], aout=['y'])
    def CopyVariable(self, x, y):
        y[...] = 1.0 * x

    @CopyVariable.defvjp
    def _(self, _y, _x):
        _x[...] = _y

    @microcode(ain=['x1', 'x2'], aout=['y'])
    def Add(self, x1, x2, y):
        if x1 is VM.Zero: y[...] = x2
        if x2 is VM.Zero: y[...] = x1
        y[...] = x1 + x2

    @Add.defvjp
    def _(self, _y, _x1, _x2):
        _x1[...] = _y
        _x2[...] = _y

    def tape(self):
        return Tape()

    def code(self):
        """ Creates a Code object for this VM.

            Build model with this.
        """
        d = {}
        t = type(self)
        for name in dir(t):
            method = getattr(t, name)
            if isinstance(method, MicroCode):
                d[name] = method
        MyCode = type("Code%s" % (type(self).__name__), (Code, ), d)
        return MyCode(self)

    def gradient(self, tape, add=None):
        """ Create a code object that backtraces the gradient of a previously
            recorded execution in `tape`.

            The `Add` microcode (None means `type(self).Add`) reduces partial
            derivatives.

        """
        newinst = self.code()

        if add is None:
            add = type(self).Add

        occurances = {}

        def emit_add(a, b, c):
            din = {}
            din[add.ain[0]] = a
            din[add.ain[1]] = b
            din[add.aout[0]] = c
            newinst.append(add, din)

        for microcode, kwargs, record in tape[::-1]:
            din = {}

            din.update(kwargs)
            for an, (uid, value) in record.items():
                din[an] = value

            # inputs
            for an in microcode.aout:
                din['_' + an] = '_' + kwargs.get(an, an)

            for an in microcode.ain:
                # need to rename to avoid accidentally overwrite
                # a finished gradient
                din['_' + an] = '#partial#_' + kwargs.get(an, an)
                uid, value = record[an]
                # print(an, '%0x' % uid, tape.get_refcount(uid), occurances.get(uid, 0))
                if tape.get_refcount(uid) == 1:
                    # direct overwriting is OK. no partials.
                    din['_' + an] = '_' + kwargs.get(an, an)
                elif occurances.get(uid, 0) == 0:
                    # first occurance, OK to overwrite?
                    # if the tape has extra operations this may overwrite a calculated gradient with
                    # a incomplete gradient.
                    din['_' + an] = '_' + kwargs.get(an, an)
                else:
                    if an in microcode.aout:
                        # rename the input for we are looking at a inplace operation.
                        emit_add("", '_' + kwargs.get(an, an), din['_' + an])

            # remove unused kwargs
            din = dict([(key, value) for key, value in din.items()
                        if key in microcode.vjp.argnames])

            newinst.append(microcode.vjp, din)
            # add partial derivatives
            for an in microcode.ain:
                uid, value = record[an]

                oc = occurances.get(uid, 0)
                occurances[uid] = oc + 1

                if tape.get_refcount(uid) > 1:
                    reduceto = '_' + kwargs.get(an, an)
                    partial = din['_' + an]
                    if oc > 0:
                        emit_add(reduceto, partial, reduceto)
                    else:
                        # move the partial to reduceto
                        # OK to use a move because a partial is only used once.
                        # thus we ignore the op if partial is already reduceto.
                        if partial != reduceto:
                            emit_add("", partial, reduceto)
                else:
                    # result already in right place
                    pass
        for uid in occurances:
            if tape._refcount[uid] != occurances[uid]:
                raise RuntimeError("FIXME: make this error more informative. Some gradients remain not fully computed.")
        inputs = newinst._find_inputs()
        for vin in inputs:
            newinst.defaults[vin] = VM.Zero
        return newinst

    @staticmethod
    def _add_to_graph(graph, init, list):
        """
            add a list of microcodes to a graph. The init node is duplicated as needed
            (because it may be used many times and mess up the diagram. It hurts to have
            very long edges like that.

        """
        source = {}
        dest = {}
        sans = {}
        for vn in init:
            source[vn] = "#INIT"
            dest[vn] = "#OUT"

        for i, record in enumerate(list):
            microcode, kwargs = record[0], record[1]

            graph.node(str(i) + str(id(microcode)), label=microcode.to_label(kwargs, html=True), )
            for an in microcode.ain:
                vn = kwargs.get(an, an)
                from_init = []
                if vn in source:
                    if vn == "": continue
                    if source[vn] == "#INIT":
                        from_init.append(vn)
                    else:
                        san = sans[vn]
                        dan = an
                        attrs = {}
                        if san != vn:
                            attrs['taillabel'] = san + ":"
                        if dan != vn:
                            attrs['headlabel'] = ":" + dan
                        attrs['label'] = vn
                        graph.edge(source[vn], str(i) + str(id(microcode)), **attrs)

                    dest[vn] = str(i) + str(id(microcode))

                if len(from_init) > 0:
                    graph.node(str(i) + "#INIT", label="<<b>#</b>>")
                    for vn in from_init:
                        graph.edge(str(i) + "#INIT", str(i) + str(id(microcode)), label=vn)

            for an in microcode.aout:
                vn = kwargs.get(an, an)
                source[vn] = str(i) + str(id(microcode))
                sans[vn] = an
                dest[vn] = "#OUT"

        # mark the potential output edges.
        for vn, node in dest.items():
            if node != "#OUT": continue
            graph.node(source[vn] + "#OUT", label="<<b>#</b>>")
            graph.edge(source[vn], source[vn] + "#OUT", label=vn)

    @microcode(ain=['x'], aout=['y'])
    def func(self, x, factor):
        """ this is a function """
        y = factor * x
        return y

    @func.defvjp
    def gfunc(self, x, factor, _y):
        _x = factor * _y
        return _x

class Tape(list):
    """ A tape records the computation of a code object.
        The tape object can then be used by the VM to build
        a gradient code object.
    """
    def __init__(self):
        list.__init__(self)
        self.init = {}
        self._refcount = {}

    def __str__(self):
        r = '-- Inputs (%08X)----\n' % id(self)
        r += '\n'.join([microcode.to_label(kwargs, d) for microcode, kwargs, d in self ])
        r += '\n'
        r += '-- Refcounts ----\n'
        r += ' '.join(["%08X : %d" % refcount for refcount in self._refcount.items()])
        return r

    def get_refcount(self, uid):
        return self._refcount.get(uid, 0)

    def record(self, microcode, kwargs, vin):
        """ add a record to the tape. Record is the argument name and the value. """
        din = {}
        for an, value in zip(microcode.argnames, vin):
            uid = id(value)
            if an in microcode.vjp.argnames:
                # need to store tha value
                din[an] = (uid, value)
            else:
                # only need the uniq id
                # can't skip the value because if it is skipped python
                # starts to recycle ids. Need a fix for this.
                din[an] = (uid, value)
            self._refcount[uid] = self._refcount.get(uid, 0) + 1
        list.append(self, (microcode, kwargs, din))

    def to_graph(self, **kwargs):
        """ create a graphviz Digraph"""
        import graphviz
        graph = graphviz.Digraph(**kwargs)
        VM._add_to_graph(graph, self.init, self)
        return graph

    def _repr_png_(self):
        return self.to_graph(engine='dot', graph_attr=dict(rankdir="LR")).pipe(format="png")

    def _repr_svg_(self):
        return self.to_graph(engine='dot', graph_attr=dict(rankdir="LR"))._repr_svg_()

class Code(list):
    """ A code object is a sequence of microcodes with input, output variables and parameters.
    """
    def __init__(self, vm):
        self.microcodes = []
        self.vm = vm
        self.defaults = {}

    def copy(self):
        r = self.vm.code()
        r.defaults.update(self.defaults)
        for microcode, kwargs in self.microcodes:
            r.append(microcode, kwargs)
        return r

    def append(self, microcode, kwargs):
        self.microcodes.append( (microcode, kwargs))

    def __repr__(self):
        r = "--Code---\n"
        r += '\n'.join(code.to_label(kwargs) for code, kwargs in self.microcodes)
        r += '\n'
        return r

    def _find_inputs(self):
        """ find inputs of a code segment """
        live = set()
        for microcode, kwargs in reversed(self.microcodes):
            for an in microcode.aout:
                vn = kwargs.get(an, an)
                if vn in live:
                    live.remove(vn)
            for an in microcode.ain:
                vn = kwargs.get(an, an)
                live.add(vn)
        return list(live)

    def _find_outputs(self):
        """ find outputs of a code segment; even if they've been used as an input """
        live = set()
        for microcode, kwargs in self.microcodes:
            for an in microcode.aout:
                vn = kwargs.get(an, an)
                live.add(vn)
        return list(live)

    def _optimize(self, vout):
        live = set(vout)
        code = self.vm.code()
        optimized = []
        for microcode, kwargs in reversed(self.microcodes):
            keep = False
            for an in microcode.aout:
                vn = kwargs.get(an, an)
                if vn in live:
                    keep = True
                    live.remove(vn)
            if keep:
                for an in microcode.ain:
                    vn = kwargs.get(an, an)
                    live.add(vn)

                optimized.append((microcode, kwargs))

        for microcode, kwargs in reversed(optimized):
            code.append(microcode, kwargs)
        return code

    def __iter__(self):
        return iter(self.microcodes)

    def compute(self, vout, init, return_tape=False, monitor=None):
        if not isinstance(vout, (tuple, list)):
            vout = [vout]
            squeeze = True
        else:
            squeeze = False

        code = self._optimize(vout)
        outputs = code._find_outputs()
        for vn in vout:
            if vn not in outputs:
                # the optimization has killed some outputs
                # the code has a bug, disable optimization to get an error message.
                code = self.copy()
                break

        inputs = code._find_inputs()

        frontier = {}

        init2 = {}
        init2.update(self.defaults)
        init2.update(init)

        for key, value in init2.items():
            # up cast 0 to VM.Zero for convenience
            if value is 0:
                value = VM.Zero
            frontier[key] = value

        frontier[""] = VM.Zero

        for vn in inputs:
            if vn not in frontier:
                raise ValueError("`%s` not defined in inputs" % vn)

        if return_tape:
            tape = Tape()
            tape.init.update(init2)
        else:
            tape = None

        started = False
        for i, (microcode, kwargs) in enumerate(code):
            try:
                r = microcode.invoke(self.vm, frontier, kwargs, tape, monitor)
            except Exception as e:
                print("Failure in running `%s`" % microcode)
                raise
            frontier.update(r)
            future = code.microcodes[i+1:]
            code._gc(frontier, future, vout, monitor)

        r = [frontier[vn] for vn in vout]
        if squeeze:
            r = r[0]
        if return_tape:
            r = r, tape
        return r

    def compute_with_gradient(self, vout, init, ginit, monitor=None):
        if not isinstance(vout, (tuple, list, set)):
            vout = [vout]
            squeeze = True
        else:
            squeeze = False

        cnout = [vn for vn in vout if not vn.startswith('_')]
        # if gradient request are requested, they must be computed
        cnout_g = [ vn[1:] for vn in ginit]

        gnout = [vn for vn in vout if vn.startswith('_')]

        cout, tape = self.compute(cnout + cnout_g, init, return_tape=True)
        cout = cout[:len(cnout)]

        gradient = self.vm.gradient(tape)

        _init = init.copy()
        _init.update(ginit)

        gout = gradient.compute(gnout, _init, monitor=monitor)
        d = {}
        d.update(zip(cnout, cout))
        d.update(zip(gnout, gout))

        out = [d[vn] for vn in vout]
        if squeeze:
            out = out[0]
        return out

    def _gc(self, frontier, future, vout, monitor=None):
        """ remove variables that are never used again """
        used = []
        used.extend(vout)
        for microcode, kwargs in future:
            for an in microcode.ain:
                vn = kwargs.get(an, an)
                used.append(vn)

        used = set(used)
        for vn in list(frontier.keys()):
            if vn not in used:
                if monitor:
                    monitor("freeing", vn)
                frontier.pop(vn)

    def to_graph(self, **kwargs):
        """ create a graphviz Digraph"""
        import graphviz
        graph = graphviz.Digraph(**kwargs)

        VM._add_to_graph(graph, self._find_inputs(), self.microcodes)

        return graph

    def _repr_png_(self):
        return self.to_graph(engine='dot', graph_attr=dict(rankdir="LR")).pipe(format="png")

    def _repr_svg_(self):
        return self.to_graph(engine='dot', graph_attr=dict(rankdir="LR"))._repr_svg_()


#####################
#
# ZeroType.
#
# Zero * anything = Zero; Zero + anything = anything
#
# VM.Zero is used to suppress gradient computation
# backtracing from VM.Zero shall always give VM.Zero
# regardless of the microcode function; this may 
# get more complicated if a microcode function
# has multiple outputs -- e.g. if the backtrace gradient
# of some are not VM.Zero.
#
# In case it occurs in a regular Python expression,
# these operator overides ensure we can 
# propagate VM.Zero properly.
#
# In other cases the microcode functions may need
# to special case on variables that are VM.Zero.

def ZeroType():
    """ creates a special type of ZeroType; """
    def self(self, *args): return self
    def other(self, other): return other
    def zde(self, a): raise ZeroDivisionError
    def __sub__(self, a): return -a
    def __xor__(self, a): return ~a
    def __int__(self): return 0
    def __float__(self): return 0.0
    def __round__(self): return 0
    def __array__(self):
        import numpy
        return numpy.array(0)
    def __repr__(self): return "<ZERO>"
    dict = {}
    for name, value in locals().items():
         if name.startswith("__"): dict[name] = value

    for name in [
        "neg", "pos", "abs", "invert", "complex",
        "mul", "rmul", "matmul", "rmatmul", 
        "mod", "divmod", "div", "truediv", "floordiv",
        "pow",
        "and", "rand", "lshift", "rlshift", "rshift", "rrshift",
        "getitem", "reversed"]:
        dict["__" + name + "__"] = self

    for name in [
        "rmod", "rdivmod", "rdiv", "rtruediv", "rfloordiv",
        "rpow", "rsub", "rxor"]:
        dict["__" + name + "__"] = zde

    for name in [
        "add", "radd", "or", "ror"]:
        dict["__" + name + "__"] = other

    dict["__repr__"] = __repr__
    return type("ZeroType", (object,), dict)

ZeroType = ZeroType()
Zero = ZeroType()

VM.Zero = Zero
VM.microcode = staticmethod(microcode)
VM.programme = staticmethod(programme)




from __future__ import print_function
from abopt.legacy.abopt import GradientDescent, LBFGS, SLBFGS
from numpy.testing import assert_raises, assert_allclose

def test_nothing():
    pass

def test_gradient_descent_customized():
    optimizer = GradientDescent(
            addmul=lambda a, b, s: a + b * s,
            dot=lambda dx1, dx2 : dx1 * dx2,
            )

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    called = [False]
    def notification(state):
        called[0] = True

    optimizer.maxsteps = 1000
    optimizer.tol = 1e-6
    optimizer.gtol = 1e-6
    optimizer.gamma = 0.001

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)
    assert called[0]
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

def test_gradient_descent():
    optimizer = GradientDescent()

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    called = [False]
    def notification(state):
        called[0] = True

    optimizer.maxsteps = 1000
    optimizer.tol = 1e-6
    optimizer.gtol = 1e-6
    optimizer.gamma = 0.001

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)
    assert called[0]
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

def test_gradient_descent_restart():
    optimizer = GradientDescent()

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    def notification(state):
        print(state)

    optimizer.maxsteps = 1000
    optimizer.tol = 1e-6
    optimizer.gtol = 1e-6
    optimizer.gamma = 0.001

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)
    result = optimizer.minimize(objective=f, gradient=df, x0=result, monitor=notification)
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

def test_lbfgs():
    optimizer = LBFGS(
            addmul=lambda a, b, s: a + b * s,
            dot=lambda dx1, dx2 : dx1 * dx2,
            )

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    called = [False]
    def notification(state):
        print(state)
        called[0] = True

    optimizer.maxsteps=1000
    optimizer.tol=1e-9
    optimizer.gtol=1e-10
    optimizer.m = 10

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)
    print(result.status)
    assert called[0]
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

def test_lbfgs_restart():
    optimizer = LBFGS()

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    def notification(state):
        print(state)

    optimizer.maxsteps=1000
    optimizer.tol=1e-5
    optimizer.gtol=1e-10
    optimizer.m = 10

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)

    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

    result = optimizer.minimize(objective=f, gradient=df, x0=result, monitor=notification)
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)


def test_lbfgs_default():
    optimizer = LBFGS()

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    called = [False]
    def notification(state):
        #print("***NOTIFICATION***")
        #print(state)
        called[0] = True

    optimizer.maxsteps=1000
    optimizer.tol=1e-5
    optimizer.gtol=1e-10
    optimizer.m = 10

    result = optimizer.minimize(objective=f, gradient=df, x0=6., monitor=notification)
    assert called[0]
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

def test_lbfgs_stochastic():
    import numpy
    optimizer = SLBFGS()
    optimizer.N0 = 10
    rng = numpy.random.RandomState(123)
    optimizer.use_linesearch = True
    optimizer.local_linesearch = False

    def df(x):
        y = 4 * x**3 - 9 * x**2
        return y
    def f(x):
        return x ** 4 - 3 * x ** 3 + 2

    called = [False]
    def notification(state):
        #print("***NOTIFICATION***")
        print(state, state.noise, state.x)
        called[0] = True

    optimizer.maxsteps=1000
    optimizer.tol=0
    optimizer.atol=1e-9
    optimizer.gtol=1e-10
    optimizer.m = 10
    optimizer.gthresh = 1e-1
    optimizer.noiseamp = 1.0
    optimizer.oracle = lambda state : rng.normal()
    result = optimizer.minimize(objective=f, gradient=df, x0=6, monitor=notification)
    assert called[0]
    assert_allclose(result['x'], 2.25, rtol=0.05)
    assert_allclose(result['g'], 0.0, atol=0.05)

    #optimizer.oracle = lambda state : 0
    #result = optimizer.minimize(objective=f, gradient=df, x0=6, monitor=notification)
    # without the oracle it falls at a stationary point.
    #assert called[0]
    #assert_allclose(result['x'], 0.0, atol=0.01)
    #assert_allclose(result['g'], 0.0, atol=0.05)



"""
    A general TrustRegion method.

    https://optimization.mccormick.northwestern.edu/index.php/Trust-region_methods

"""

from abopt.base import Optimizer, Problem, Proposal
from abopt.base import ContinueIteration, ConvergedIteration, FailedIteration
from abopt.linesearch import backtrace

class TrustRegionCG(Optimizer):
    optimizer_defaults = {'eta1' : 0.1,
                        'eta2' : 0.25,
                        'eta3' : 0.75,
                        't1' : 0.25,
                        't2' : 2.0,
                        'maxiter' : 1000,
                        'm' : 6,
                        'conviter' : 6,
                        'cg_preconditioner' : None,
                        'linesearch' : backtrace,
                        'linesearchiter' : 100,
                        'cg_monitor' : None,
                        'cg_maxiter' : 50,
                        'cg_rtol' : 1e-2,
                        'maxradius' : 100.,
                        'minradius' : 1e-9,
                        'initradius' : None,
                        }

    def single_iteration(self, problem, state):
        mul = problem.vs.mul
        dot = problem.vs.dot
        addmul = problem.vs.addmul

        def Avp(v):
            state.hev = state.hev + 1
            return problem.PHvp(state.x, v)

        def cg_monitor(*kwargs):
            if self.cg_monitor is not None:
                self.cg_monitor(*kwargs)

        if self.cg_preconditioner:
            C = self.cg_preconditioner(Avp)
        else:
            C = None

        # solve - H z = g constrained by the radius
        radius1 = state.radius

        z = cg_steihaug(problem.vs, Avp, state.Pg, state.z, radius1,
                self.cg_rtol, self.cg_maxiter, monitor=cg_monitor, C=C)

        mdiff = 0.5 * dot(z, Avp(z)) - dot(state.Pg, z)

        Px1 = addmul(state.Px, z, -1)
        x1 = problem.Px2x(Px1)
        y1 = problem.f(x1)
        state.fev = state.fev + 1

        fdiff = y1 - state.y

        if mdiff < 0:
            rho = fdiff / mdiff
        else:
            rho = 0

#        print(y1, x1)
#        print(state.y, state.x)
        #print('rho', rho, 'fdiff', fdiff, 'mdiff', mdiff, 'Avp(z)', Avp(z), 'Pg', state.Pg, 'znorm', dot(z, z) ** 0.5, 'radius', radius1)

        interior = dot(z, z) ** 0.5 < 0.9 * radius1

        if rho < self.eta1:
            # poor descent stay and shrink
            # restart from the previus cg_steihaug result, but shrink the size to avoid
            # excessive recoveries.
            radius1 = min(self.t1 * radius1, state.Pgnorm)
            prop = Proposal(problem, Px=state.Px, x=state.x, y=state.y, z=mul(z, 0.9 * radius1 / state.radius))
            prop.message = "poor descent "
        elif rho < self.eta2:
            # poor approximation but good descent, move and shrink
            radius1 = min(self.t1 * radius1, state.Pgnorm)
            prop = Proposal(problem, Px=Px1, x=x1, y=y1, z=mul(z, 0.9 * radius1 / state.radius))
            prop.message = "poor approximation "
            # reinialize radius from the gradient norm if needed
        elif rho > self.eta3 and not interior:
            # good and too conservative, move and grow
            radius1 = min(radius1 * self.t2, self.maxradius)
            prop = Proposal(problem, Px=Px1, x=x1, y=y1, z=z)
            prop.message = "too conservative"
        else: # about right
            # good and too conservative, move and grow
            radius1 = radius1
            prop = Proposal(problem, Px=Px1, x=x1, y=y1, z=z)
            prop.message = "good approximation"


        prop.radius = radius1
        prop.rho = rho

        return prop

    def assess(self, problem, state, prop):
        if prop is None:
            return FailedIteration("no proposal is made")

        prop = prop.complete(state)

        #print("assess radius", state.radius, 'tol', problem.get_tol(state.y), 'gnorm', prop.gnorm, 'gtol', problem.gtol)
        if prop.radius >= state.radius:
            if problem.check_convergence(state.y, prop.y):
                return ConvergedIteration("Objective is not improving in trust region")
            if prop.dxnorm <= problem.xtol:
                return ConvergedIteration("Solution is not moving in trust region")

        if prop.radius < self.minradius:
            return ConvergedIteration("Trust region is too small.")

        if prop.gnorm <= problem.gtol:
            return ConvergedIteration("Gradient is sufficiently small")

        return ContinueIteration("normal iteration")

    def start(self, problem, state, x0):
        prop = Optimizer.start(self, problem, state, x0)

        if self.initradius is None:
            prop.radius = min(prop.Pgnorm, self.maxradius)
        else:
            prop.radius = self.initradius

        prop.rho = 1.0
        return prop

    def move(self, problem, state, prop):
        state.radius = prop.radius
        state.rho = prop.rho

        #print('move', prop.y)
        Optimizer.move(self, problem, state, prop)

def cg_steihaug(vs, Avp, g, z0, Delta, rtol, maxiter=1000, monitor=None, C=None):
    """ best effort solving for y = A^{-1} g with cg,
        given the trust-region constraint;

        This is roughly ported from Jeff Regier's

            https://github.com/jeff-regier/Celeste.jl/blob/316cae12efcc394a5253ee799056c5513c2efc67/src/cg_trust_region.jl

        the algorithm is almost identical to the one on NWU wiki,

            https://optimization.mccormick.northwestern.edu/index.php/Trust-region_methods

        if given B, update approximated Hessian, see Morale and Nocedal, 2000

            http://epubs.siam.org/doi/abs/10.1137/S1052623497327854


        C(v, direction) the preconditioner operator. C^{direction}. 
        Usually, C is close to A, such that C^{-1} A has a lower condition number.

        One particular case is if C is the diagonal of A.

        See Steihaug's paper. https://epubs.siam.org/doi/pdf/10.1137/0720042

    """
    if C is None: C = lambda x, direction: x

    dot = vs.dot
    mul = vs.mul
    addmul = vs.addmul

    if z0 is None:
        z0 = vs.zeros_like(g)

    # FIXME: how to seed cg with a different starting point?
    r0 = addmul(Avp(z0), g, -1)
    mr0 = C(r0, -1)
    d0 = mul(mr0, 1)

    j = 0

    rho_init = dot(mr0, r0)   # <r, mr>

    rho0 = rho_init

    while True:
        Bd0 = Avp(d0)
        dBd0 = dot(d0, Bd0)  # gamma

        alpha = rho0 / dBd0

        p0 = addmul(z0, d0, -alpha)

        message = ""

        if dBd0 == 0: # zero Hessian
            rho1 = 0 # will terminate
            z1 = z0
            r1 = r0
            mr1 = mr0
            d1 = d0
            message = "zero hessian"

        elif dBd0 <= 0 or dot(p0, C(p0, 1)) ** 0.5 >= Delta:
            #print("dBd0", dBd0, "rad", dot(p0, p0) ** 0.5, Delta)
            # negative curvature or too fast
            # find tau such that p = z0 + tau d0 minimizes m(p)
            # and satisfies ||pk|| == \Delta_k.
            a_ = dot(d0, C(d0, 1))
            b_ = 2 * dot(z0, C(d0, 1))
            c_ = dot(z0, C(z0, 1)) - Delta ** 2
            cond = (b_ **2 - 4 * a_ * c_)
            if a_ == 0 or cond < 0: # already at the solution, do not move.
                rho1 = -1
                tau = 1.0
                if c_ > 0:
                    tau = Delta / c_ ** 0.5
                z1 = mul(z0, tau)
                if a_ == 0:
                    message = "already at the right direction"
                    rho1 = -1
                else:
                    message = "no solution to second order equation, restarting "
                    z0 = vs.zeros_like(g)
                    r0 = addmul(Avp(z0), g, -1)
                    mr0 = C(r0, -1)
                    d0 = mul(mr0, 1)
                    rho1 = dot(mr0, r0)   # <r, mr>
            else:
                tau = (- b_ - cond ** 0.5) / (2 * a_)
                # do not assert because when d0 is close to zero
                # tau may be a large number
                # assert tau <= 0
                z1 = addmul(z0, d0, tau)

                if dBd0 <= 0:
                    rho1 = -1 # will terminate
                else:
                    rho1 = 0
                if dBd0 <= 0:
                    message = "negative curvature "
                else:
                    message = "truncation"

            r1 = r0
            mr1 = mr0
            d1 = d0
        else:
            z1 = addmul(z0, d0,  -alpha)
            r1 = addmul(r0, Bd0, -alpha)
            mr1 = C(r1, -1)

            rho1 = dot(mr1, r1)
            d1 = mul(mr1, 1)
            d1 = addmul(d1, d0, rho1 / rho0)


            message = "regular iteration"

        r0 = r1
        mr0 = mr1
        d0 = d1
        z0 = z1
        rho0 = rho1

        if monitor is not None:
#            monitor(j, rho0, r0, d0, z0, Avp(z0), g, B)
            z0norm = dot(z0, z0)
            zgnorm = dot(z0, g)
            gnorm = dot(g, g)
            monitor(j, message, rho0, rho_init, rtol, zgnorm / z0norm ** 0.5 / gnorm ** 0.5)

        if rho1 / rho_init < rtol ** 2:
            #print("rho1 / rho_init", rho1 / rho_init, rtol ** 2)
            break

        if j >= maxiter:
            break
        j = j + 1

    return z0



# expose optimization algorithms
from .algs.lbfgs import LBFGS
from .algs.gradient import GradientDescent
from .algs.newton import DirectNewton
from .algs.trustregion import TrustRegionCG

# expose common vector spaces
from .vectorspace import real_vector_space
from .vectorspace import complex_vector_space

# providing base classes here
# for external subclassing

from .base import VectorSpace
from .base import State
from .base import Preconditioner
from .base import Problem

def minimize(optimizer, objective, gradient, x0, hessian_vector_product=None,
    monitor=None, vs=real_vector_space, precond=None):

    problem = Problem(objective, gradient, hessian_vector_product=hessian_vector_product, vs=vs, precond=precond)

    return optimizer.minimize(problem, x0, monitor=monitor)



from abopt.base import Optimizer
from abopt.linesearch import backtrace

class DirectNewton(Optimizer):
    """
        A direct Newton method.

        This method assumes the problem knows how to compute the inverse of hessian,
        without actually inverting it.

        A line search along the newton proposal direction, starting from the newton step size
        is used.
    """

    optimizer_defaults = {
                        'maxiter' : 1000,
                        'conviter' : 2,
                        'linesearch' : backtrace,
                        'linesearchiter' : 100,
                        }

    def single_iteration(self, problem, state):
        mul = problem.vs.mul
        dot = problem.vs.dot
        addmul = problem.vs.addmul

        state.hev = state.hev + 1
        z = problem.Phvp(state.x, state.Pg)

        prop, r1 = self.linesearch(problem, state, z, 1.0, maxiter=self.linesearchiter)

        return prop
